from snudda.neurons.neuron_morphology import NeuronMorphology
from snudda.neurons.neuron_morphology_extended import NeuronMorphologyExtended
from snudda.neurons.neuron_model_extended import NeuronModel

# This depends on NEURON -- so let's not have it as a mandatory import
# from snudda.neurons.neuron_modulation import NeuronModulation



import numpy as np

class NEURONNeuronExtractor:

    def __init__(self, sim):

        self.sim = sim

    def extract_geometry_for_neuron(self, neuron_id):

        loc_info = []

        for sec_id, sec in self.sim.neurons[neuron_id].section_lookup.items():

            if sec.n3d() <= 1:
                continue

            coords = np.array([[sec.x3d(i), sec.y3d(i), sec.z3d(i), sec.diam3d(i), sec.arc3d(i)]
                                for i in range(sec.n3d())])

            arc_points = np.linspace(0.0,1.0, num=sec.nseg+1, endpoint=True)

            xp = np.interp(arc_points, coords[:, 4], coords[:,0])
            yp = np.interp(arc_points, coords[:, 4], coords[:,1])
            zp = np.interp(arc_points, coords[:, 4], coords[:,2])
            dp = np.interp(arc_points, coords[:, 4], coords[:,3])

            for i in range(0, sec.nseg):
                loc_info.append(np.array([xp[i], xp[i+1],
                                          yp[i], yp[i+1],
                                          zp[i], zp[i+1],
                                          dp[i], dp[i+1]]))

        geometry = np.vstack(loc_info) * 1e-6  # We save in meter

        return geometry

    def extract_geometry_for_neuron_OLD(self, neuron_id):

        loc_info = []

        for sec_id, sec in self.sim.neurons[neuron_id].section_lookup.items():

            if sec.n3d() > 1:
                loc_info.append(np.array([[sec.x3d(i), sec.x3d(i+1),
                                           sec.y3d(i), sec.y3d(i+1),
                                           sec.z3d(i), sec.z3d(i+1),
                                           sec.diam3d(i), sec.diam3d(i+1)]
                                          for i in range(sec.n3d()-1)]))

        geometry = np.vstack(loc_info) * 1e-6  # We save in meter

        return geometry

    def extract_coordinates_for_neuron(self, neuron_id):

        loc_info = []

        for sec_id, sec in self.sim.neurons[neuron_id].section_lookup.items():

            if sec.n3d() > 1:
                loc_info.append(np.array([[sec.x3d(i), sec.y3d(i), sec.z3d(i), sec.diam3d(i), sec.arc3d(i)]
                                          for i in range(sec.n3d())]))

        coordinates = np.vstack(loc_info) * 1e-6  # We save in meter

        return coordinates

    def write_all_coordinates_to_hdf5(self, hdf5_file):

        print(f"Writing geometry to {hdf5_file.file.filename}")

        for neuron_id in self.sim.neurons.keys():

            geometry = self.extract_geometry_for_neuron(neuron_id=neuron_id)
            hdf5_file["neurons"][str(neuron_id)].create_dataset("geometry", data=geometry)




import os
import subprocess
import time
import unittest


class TestProjectionDetection(unittest.TestCase):

    def setUp(self):

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

        from snudda.place.create_cube_mesh import create_cube_mesh

        # Create cube meshes
        self.network_path = os.path.join("networks", "network_testing_projection_detection")
        mesh_file_a = os.path.join(self.network_path, "mesh", "volume_A.obj")
        mesh_file_b = os.path.join(self.network_path, "mesh", "volume_B.obj")

        create_cube_mesh(mesh_file_a, [5e-3, 0, 0], 300e-6, "Volume A - connect structures example")
        create_cube_mesh(mesh_file_b, [-5e-3, 0, 0], 300e-6, "Volume B - connect structures example")

        # Define network
        from snudda.init.init import SnuddaInit

        cnc = SnuddaInit(network_path=self.network_path, random_seed=123)
        cnc.define_structure(struct_name="VolumeA", struct_mesh=mesh_file_a, d_min=15e-6, mesh_bin_width=50e-6)
        cnc.define_structure(struct_name="VolumeB", struct_mesh=mesh_file_b, d_min=15e-6, mesh_bin_width=50e-6)

        cnc.add_neurons(name="dSPN", num_neurons=20, volume_id="VolumeA",
                        neuron_dir=os.path.join("$DATA", "neurons", "striatum", "dspn"))
        cnc.add_neurons(name="iSPN", num_neurons=20, volume_id="VolumeB",
                        neuron_dir=os.path.join("$DATA", "neurons", "striatum", "ispn"))

        # Add the projection we want to test dSPN->iSPN
        proj_file = os.path.join("data", "ExampleProjectionDetection.json")

        cnc.add_neuron_target(neuron_name="dSPN",
                              target_name="iSPN",
                              region_name="VolumeA",
                              connection_type="GABA",
                              dist_pruning=None,
                              f1=None,
                              soft_max=None,
                              mu2=None,
                              a3=None,
                              conductance=5e-9,
                              cluster_synapses=False,
                              mod_file="tmGlutA",
                              parameter_file=None,
                              channel_param_dictionary=None,
                              projection_file=proj_file)

        # Also add dSPN-dSPN and iSPN-iSPN synapses
        # Note we do NOT add dSPN-iSPN again this way, as that would overwrite the above connections
        # (The above neuron_projection will also do normal touch detection)

        SPN2SPNdistDepPruning = "1-exp(-(0.4*d/60e-6)**2)"

        MSD1gGABA = [0.24e-9, 0.1e-9]
        MSD2gGABA = [0.24e-9, 0.1e-9]

        MSD1GABAfailRate = 0.7  # Taverna 2008, figure 2
        MSD2GABAfailRate = 0.4  # Taverna 2008, 2mM

        pfdSPNdSPN = os.path.join("$DATA", "synapses", "striatum", "PlanertFitting-DD-tmgaba-fit.json")
        pfdSPNiSPN = os.path.join("$DATA", "synapses", "striatum", "PlanertFitting-DI-tmgaba-fit.json")
        pfiSPNdSPN = os.path.join("$DATA", "synapses", "striatum", "PlanertFitting-ID-tmgaba-fit.json")
        pfiSPNiSPN = os.path.join("$DATA", "synapses", "striatum", "PlanertFitting-II-tmgaba-fit.json")

        cnc.add_neuron_target(neuron_name="dSPN",
                              target_name="dSPN",
                              region_name="VolumeA",
                              connection_type="GABA",
                              dist_pruning=SPN2SPNdistDepPruning,
                              f1=0.38, soft_max=3, mu2=2.4,
                              a3=1.0,
                              conductance=MSD1gGABA,
                              parameter_file=pfdSPNdSPN,
                              mod_file="tmGabaA",
                              channel_param_dictionary={"tau1": (1.3e-3, 1e3),
                                                        "tau2": (12.4e-3, 1e3),
                                                        "failRate": MSD1GABAfailRate})

        cnc.add_neuron_target(neuron_name="iSPN",
                              target_name="iSPN",
                              region_name="VolumeB",
                              connection_type="GABA",
                              dist_pruning=SPN2SPNdistDepPruning,
                              f1=0.55, soft_max=4, mu2=2.4,
                              a3=1.0,
                              conductance=MSD2gGABA,
                              parameter_file=pfiSPNiSPN,
                              mod_file="tmGabaA",
                              channel_param_dictionary={"tau1": (1.3e-3, 1e3),
                                                        "tau2": (12.4e-3, 1e3),
                                                        "failRate": MSD2GABAfailRate})

        cnc.write_json()

        # Place neurons, then detect, project and prune

        from snudda.place.place import SnuddaPlace
        sp = SnuddaPlace(network_path=self.network_path, verbose=True)
        sp.parse_config()
        sp.write_data()

        from snudda.detect.detect import SnuddaDetect
        sd = SnuddaDetect(network_path=self.network_path, hyper_voxel_size=100, verbose=True)
        sd.detect()

        # Old version of projection
        from snudda.detect.project import SnuddaProject
        sp = SnuddaProject(network_path=self.network_path)
        sp.project()

        from snudda.detect.prune import SnuddaPrune
        sp = SnuddaPrune(network_path=self.network_path, verbose=True)
        sp.prune()

    def test_project(self):

        # Are there connections dSPN->iSPN
        from snudda.utils.load import SnuddaLoad
        network_file = os.path.join(self.network_path, "network-synapses.hdf5")
        sl = SnuddaLoad(network_file)

        dspn_id_list = sl.get_neuron_id_of_type("dSPN")
        ispn_id_list = sl.get_neuron_id_of_type("iSPN")

        tot_proj_ctr = 0

        for dspn_id in dspn_id_list:
            for ispn_id in ispn_id_list:

                synapses, synapse_coords = sl.find_synapses(pre_id=dspn_id, post_id=ispn_id)
                if synapses is not None:
                    tot_proj_ctr += synapses.shape[0]

        with self.subTest(stage="projection_exists"):
            # There should be projection synapses between dSPN and iSPN in this toy example
            self.assertTrue(tot_proj_ctr > 0)

        tot_dd_syn_ctr = 0
        for dspn_id in dspn_id_list:
            for dspn_id2 in dspn_id_list:

                synapses, synapse_coords = sl.find_synapses(pre_id=dspn_id, post_id=dspn_id2)
                if synapses is not None:
                    tot_dd_syn_ctr += synapses.shape[0]

        tot_ii_syn_ctr = 0
        for ispn_id in ispn_id_list:
            for ispn_id2 in ispn_id_list:

                synapses, synapse_coords = sl.find_synapses(pre_id=ispn_id, post_id=ispn_id2)
                if synapses is not None:
                    tot_ii_syn_ctr += synapses.shape[0]

        with self.subTest(stage="normal_synapses_exist"):
            # In this toy example neurons are quite sparsely placed, but we should have at least some
            # synapses
            self.assertTrue(tot_dd_syn_ctr > 0)
            self.assertTrue(tot_ii_syn_ctr > 0)

        # We need to run in parallel also to verify we get same result (same random seed)

        serial_synapses = sl.data["synapses"].copy()
        del sl  # Close old file so we can overwrite it

        os.environ["IPYTHONDIR"] = os.path.join(os.path.abspath(os.getcwd()), ".ipython")
        os.environ["IPYTHON_PROFILE"] = "default"
        # os.system("ipcluster start -n 4 --profile=$IPYTHON_PROFILE --ip=127.0.0.1&")
        self.cluster_process = subprocess.Popen(
            ["ipcluster", "start", "-n", "4", "--profile=default", "--ip=127.0.0.1"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

        time.sleep(3)

        from ipyparallel import Client
        u_file = os.path.join(".ipython", "profile_default", "security", "ipcontroller-client.json")

        # Wait for connection file to exist
        for _ in range(30):
            if os.path.exists(u_file):
                break
            time.sleep(1)
            print(".")
        else:
            raise RuntimeError("ipcontroller connection file not created")

        print(f"Found {u_file}, proceeding.")

        rc = Client(url_file=u_file, timeout=120, debug=False)

        try:
            rc.wait_for_engines(n=4, timeout=60)
        except Exception as e:
            raise RuntimeError(f"Engines did not start within 60 seconds: {e}")

        #################

        d_view = rc.direct_view(targets='all')  # rc[:] # Direct view into clients

        from snudda.detect.detect import SnuddaDetect
        sd = SnuddaDetect(network_path=self.network_path, hyper_voxel_size=100, rc=rc, verbose=True)
        sd.detect()

        # Old version of project, the new version with touch detection is in detect.
        from snudda.detect.project import SnuddaProject
        # TODO: Currently SnuddaProject only runs in serial
        sp = SnuddaProject(network_path=self.network_path)
        sp.project()

        from snudda.detect.prune import SnuddaPrune
        # Prune has different methods for serial and parallel execution, important to test it!
        sp = SnuddaPrune(network_path=self.network_path, rc=rc, verbose=True)
        sp.prune()

        sd = None
        sp = None

        with self.subTest(stage="check-parallel-identical"):
            sl2 = SnuddaLoad(network_file)
            parallel_synapses = sl2.data["synapses"].copy()

            # ParameterID, sec_X etc are randomised in hyper voxel, so you need to use same
            # hypervoxel size for reproducability between serial and parallel execution

            print(f"serial shape: {serial_synapses.shape}, parallel shape: {parallel_synapses.shape}")

            if serial_synapses.shape != parallel_synapses.shape:
                import json
                from snudda.utils.numpy_encoder import NumpyEncoder

                save_problem = dict()
                save_problem["serial"] = serial_synapses
                save_problem["parallel"] = parallel_synapses

                with open("projection-detection-synapses-mismatch.json", "wt") as f:
                    json.dump(save_problem, f, indent=4, cls=NumpyEncoder)

            # All synapses should be identical regardless of serial or parallel execution path
            self.assertTrue(serial_synapses.shape == parallel_synapses.shape,
                            f"serial shape: {serial_synapses.shape}, parallel shape: {parallel_synapses.shape}\n"
                            f"Serial synapse matrix: {serial_synapses}\n"
                            f"Parallel synapse matrix: {parallel_synapses}\n")
            self.assertTrue((serial_synapses == parallel_synapses).all())

        # os.system("ipcluster stop")

        if self.cluster_process.stdout:
            self.cluster_process.stdout.close()
        if self.cluster_process.stderr:
            self.cluster_process.stderr.close()

        rc.shutdown(hub=True)
        rc.close()

        # self.cluster_process.terminate()  # sends SIGTERM
        # try:
        #     self.cluster_process.wait(timeout=10)
        # except subprocess.TimeoutExpired:
        #     self.cluster_process.kill()


if __name__ == '__main__':
    unittest.main()



# This script reads an SBML file, and creates a snudda neuromodulation file
import os
import json
import libsbml   # If missing, use:  pip install python-libsbml
import numpy as np


class ReadSBML:

    def __init__(self, filename=None, out_file=None, concentration_scale_factor=None):

        self.filename = filename
        self.out_file = out_file
        self.reader = None
        self.data = None

        if concentration_scale_factor is None:
            self.concentration_scale_factor = 1
        else:
            self.concentration_scale_factor = concentration_scale_factor
            print(f"Using concentration scale factor: {self.concentration_scale_factor}")

        if filename is not None:

            if not os.path.isfile(filename):
                raise ValueError(f"File not found {filename}")
            self.parse()

            if out_file is not None:
                self.write(out_file=out_file)

    def parse(self):

        print(f"Reading {self.filename}")

        self.reader = libsbml.SBMLReader()
        document = self.reader.readSBML(self.filename)

        model = document.getModel()

        species_data = {}
        id_to_name = {}

        # Extract global parameters
        global_parameters = {}
        for param in model.getListOfParameters():
            global_parameters[param.getId()] = param.getValue()

        for species in model.getListOfSpecies():
            species_name = species.getName().replace('*', '_')
            species_id = species.getId()

            if species_id in id_to_name:
                raise KeyError(f"{species_id =} already defined.")
            id_to_name[species_id] = species_name

            initial_concentration = species.getInitialConcentration()
            charge = species.getCharge()
            initial_concentration = species.getInitialConcentration()
            diffusion_constant = 0  # Assuming a default value as it's not in SBML
            regions = ["soma_internal", "dend_internal"]  # Assuming these regions
            atol_scale = None  # Assuming a default value
            boundary_condition = species.getBoundaryCondition()  # Assuming a default value
            # represents = species.getId()  # Assuming species ID represents itself

            species_data[species_name] = {
                "initial_concentration": initial_concentration * self.concentration_scale_factor,
                "diffusion_constant": diffusion_constant,
                "charge": charge,
                "regions": regions,
                "concentration": initial_concentration * self.concentration_scale_factor
                # "represents": represents
            }

            if atol_scale is not None:
                species_data[species_name]["atol_scale"] = atol_scale

            if boundary_condition is not None:
                species_data[species_name]["boundary_condition"] = boundary_condition

        # Using this viewer: https://sv.insysbio.com/online/
        # We have identified that the names with "*" in them are forward and backward rates
        # we need to extract those, and also extract the reaction

        # Extract reactions information
        reactions_data = {}
        for reaction in model.getListOfReactions():
            reaction_id = reaction.getName()
            reactants = " + ".join([f"{self.get_stoichiometry(reactant)} * {id_to_name[reactant.getSpecies()]}"
                                    if reactant.getStoichiometry() != 1 else id_to_name[reactant.getSpecies()]
                                   for reactant in reaction.getListOfReactants()])
            products = " + ".join([f"{self.get_stoichiometry(product)} * {id_to_name[product.getSpecies()]}"
                                   if product.getStoichiometry() != 1 else id_to_name[product.getSpecies()]
                                   for product in reaction.getListOfProducts()])

            forward_rate, backward_rate = self.extract_rates(reaction, model, global_parameters)

            reactions_data[reaction_id] = {
                "reactants": reactants,
                "products": products,
                "forward_rate": forward_rate,
                "backward_rate": backward_rate,
                "regions": regions
            }

        # Combine data into the desired format
        self.data = {
            "species": species_data,
            "reactions": reactions_data
        }

    def get_stoichiometry(self, reactant):
        factor = reactant.getStoichiometry()

        factor_int = int(np.round(factor))

        if np.abs(factor_int - factor) > 1e-8:
            raise ValueError(f"Error, rounding reactant incorrectly ({reactant}): {factor} -> {factor_int}")

        return factor_int

    def extract_rates(self, reaction, model, global_parameters):

        # TODO: Check that we can handle cAMP ** 2

        compartment_list = [x.getId() for x in model.getListOfCompartments()]
        species_list = [x.getId() for x in model.getListOfSpecies()]

        print(f"{reaction.getKineticLaw().getFormula()}")

        # This function takes a reaction and attempts to extract the forward and backward rates
        current_expression = reaction.getKineticLaw().getMath()
        operator = current_expression.getOperatorName()

        if operator == "times" and current_expression.getLeftChild().getName() in compartment_list:
            print(f"Removing volume {current_expression.getLeftChild().getName()}")
            current_expression = current_expression.getRightChild()

        if operator == "times" and current_expression.getRightChild().getName() in compartment_list:
            print(f"Removing volume {current_expression.getRightChild().getName()}")
            current_expression = current_expression.getLeftChild()

        operator = current_expression.getOperatorName()

        if operator == "minus":
            # We have both forward and backward rate to extract
            forward_rate = self._get_rate_helper(current_expression.getLeftChild(),
                                                 global_parameters, compartment_list, model)
            backward_rate = self._get_rate_helper(current_expression.getRightChild(),
                                                  global_parameters, compartment_list, model)

        else:
            forward_rate = self._get_rate_helper(current_expression,
                                                 global_parameters, compartment_list, model)
            backward_rate = None

        print(f"{forward_rate =}, {backward_rate =}")

        return forward_rate, backward_rate

    def _count_reactants(self, expression, model):

        num_species = 0
        species_list = [x.getId() for x in model.getListOfSpecies()]

        if expression.getId() in species_list or expression.getName() in species_list:
            # We found a species, number of reactants is 1
            return 1

        if expression.getOperatorName() == "times":
            left_child = expression.getLeftChild()
            right_child = expression.getRightChild()

            num_species += self._count_reactants(left_child, model)
            num_species += self._count_reactants(right_child, model)
        elif expression.getName() == "power":
            # libsbml.formulaToString(orig_expression)
            left_child = expression.getLeftChild()
            right_child = expression.getRightChild()

            base = self._count_reactants(left_child, model)
            exponent = right_child.getValue()

            num_species = base * exponent

        elif expression.getOperatorName() is None:
            return 0
        else:
            import pdb
            pdb.set_trace()
            raise ValueError(f"Not a multiplication: {expression.getOperatorName()}")

        return num_species

    def _get_rate_helper(self, expression, global_parameters, compartment_list, model):

        orig_expression = expression

        # TODO: We need to rescale the rates with self.concentration_scale_factor for two-factors
        #       and self.concentration_scale_factor**2 for three-factors

        # TODO: How to count reactants?

        while expression.getOperatorName() == "times":
            left_child = expression.getLeftChild()

            # If the left child is a volume (ie has the name of a volume in compartment_list),
            # then take the right child instead, and end the loop
            if left_child.getName() in compartment_list:
                expression = expression.getRightChild()
                break

            print(f"Removing: {expression.getRightChild().getName()}")
            expression = left_child

        rate_name = expression.getName()

        print(f"rate_name = {rate_name}")

        try:
            rate = global_parameters[rate_name]
        except:
            import traceback
            print(traceback.format_exc())
            import pdb
            pdb.set_trace()

        if self.concentration_scale_factor != 1:
            num_species = self._count_reactants(expression=orig_expression, model=model)

            rate /= self.concentration_scale_factor ** (num_species - 1)

        return rate

    def write(self, out_file=None):

        if out_file is None:
            out_file = self.out_file

        if out_file is None:
            raise ValueError(f"No outfile specified.")

        print(f"Writing JSON to {out_file}")

        with open(out_file, "wt") as f:
            json.dump(self.data, f, indent=4)


def cli():

    import argparse

    parser = argparse.ArgumentParser(description="Convert SBML to Snudda reaction diffusion with RxD",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("sbml_file", help="Input SBML file to convert")
    parser.add_argument("json_file", help="Snudda RxD JSON output file")
    parser.add_argument("--conc_scale_factor", default=1, help="Rescale concentrations by factor", type=float)

    args = parser.parse_args()

    rs = ReadSBML(filename=args.sbml_file, out_file=args.json_file, concentration_scale_factor=args.conc_scale_factor)


if __name__ == "__main__":

    cli()




import unittest
import numpy as np
from snudda.input import SnuddaInput


class MyTestCase(unittest.TestCase):

    def setUp(self) -> None:

        self.input = SnuddaInput()
        self.rng = np.random.default_rng(113)

    def test_lognormal_basic_functionality(self):
        """Test basic spike generation with single time range"""
        mean_freq = np.array([10.0])  # 10 Hz
        std_freq = np.array([2.0])  # 2 Hz std
        time_ranges = (np.array([0.0]), np.array([1.0]))  # 1 second

        spikes = self.input.generate_log_normal_spikes(mean_freq, std_freq, time_ranges, self.rng)

        # Check output properties
        self.assertIsInstance(spikes, np.ndarray)
        self.assertTrue(len(spikes) > 0, "Should generate some spikes")
        self.assertTrue(np.all(spikes >= 0.0), "All spikes should be >= start time")
        self.assertTrue(np.all(spikes < 1.0), "All spikes should be < end time")
        self.assertTrue(np.all(np.diff(spikes) > 0), "Spikes should be sorted")

    def test_lognormal_multiple_time_ranges(self):
        """Test with multiple time ranges"""
        mean_freq = np.array([5.0, 15.0])
        std_freq = np.array([1.0, 3.0])
        time_ranges = (np.array([0.0, 2.0]), np.array([1.0, 3.0]))

        spikes = self.input.generate_log_normal_spikes(mean_freq, std_freq, time_ranges, self.rng)

        # Check that spikes exist in both ranges
        spikes_range1 = spikes[(spikes >= 0.0) & (spikes < 1.0)]
        spikes_range2 = spikes[(spikes >= 2.0) & (spikes < 3.0)]

        no_spikes = spikes[(spikes > 1.0) & (spikes < 2.0)]

        self.assertTrue(len(spikes_range1) > 0, "Should have spikes in first range")
        self.assertTrue(len(spikes_range2) > 0, "Should have spikes in second range")
        self.assertTrue(np.all(np.diff(spikes) > 0), "All spikes should be sorted")
        self.assertTrue(len(no_spikes) == 0, "Middle part should be empty")

    def test_lognormal_empty_time_range(self):
        """Test with zero duration time range"""
        mean_freq = np.array([10.0])
        std_freq = np.array([2.0])
        time_ranges = (np.array([1.0]), np.array([1.0]))  # Zero duration

        spikes = self.input.generate_log_normal_spikes(mean_freq, std_freq, time_ranges, self.rng)

        self.assertEqual(len(spikes), 0, "Should generate no spikes for zero duration")

    def test_negative_duration(self):
        """Test with negative duration (end < start)"""
        mean_freq = np.array([10.0])
        std_freq = np.array([2.0])
        time_ranges = (np.array([2.0]), np.array([1.0]))  # Negative duration

        spikes = self.input.generate_log_normal_spikes(mean_freq, std_freq, time_ranges, self.rng)

        self.assertEqual(len(spikes), 0, "Should generate no spikes for negative duration")

    def test_high_frequency(self):
        """Test with high frequency to ensure many spikes are generated"""
        mean_freq = np.array([100.0])  # 100 Hz
        std_freq = np.array([20.0])
        time_ranges = (np.array([0.0]), np.array([1.0]))

        spikes = self.input.generate_log_normal_spikes(
            mean_freq, std_freq, time_ranges, self.rng
        )

        # Should have roughly 100 spikes (give or take due to randomness)
        self.assertGreater(len(spikes), 50, "Should generate many spikes at high frequency")
        self.assertLess(len(spikes), 200, "Shouldn't generate too many spikes")

    def test_low_frequency(self):
        """Test with very low frequency"""
        mean_freq = np.array([0.5])  # 0.5 Hz (one spike every 2 seconds on average)
        std_freq = np.array([0.1])
        time_ranges = (np.array([0.0]), np.array([1.0]))  # 1 second window

        spikes = self.input.generate_log_normal_spikes(
            mean_freq, std_freq, time_ranges, self.rng
        )

        # Might have 0 or 1 spike in 1 second window
        self.assertLessEqual(len(spikes), 3, "Should generate few spikes at low frequency")

    def test_reproducibility(self):
        """Test that same seed produces same results"""
        mean_freq = np.array([10.0])
        std_freq = np.array([2.0])
        time_ranges = (np.array([0.0]), np.array([1.0]))

        # Generate spikes twice with same seed
        rng1 = np.random.RandomState(123)
        rng2 = np.random.RandomState(123)

        spikes1 = self.input.generate_log_normal_spikes(mean_freq, std_freq, time_ranges, rng1)
        spikes2 = self.input.generate_log_normal_spikes(mean_freq, std_freq, time_ranges, rng2)

        np.testing.assert_array_equal(spikes1, spikes2, "Same seed should produce same spikes")

    def test_frequency_approximation(self):
        """Test that generated spike rate approximately matches input frequency"""
        mean_freq = np.array([20.0])  # 20 Hz
        std_freq = np.array([4.0])
        time_ranges = (np.array([0.0]), np.array([10.0]))  # Long duration for better statistics

        spikes = self.input.generate_log_normal_spikes(
            mean_freq, std_freq, time_ranges, self.rng
        )

        # Calculate actual frequency
        duration = 10.0
        actual_freq = len(spikes) / duration

        # Should be within reasonable range (allow for randomness)
        self.assertGreater(actual_freq, 15.0, "Frequency should be roughly correct (lower bound)")
        self.assertLess(actual_freq, 30.0, "Frequency should be roughly correct (upper bound)")

    def test_spike_time_bounds(self):
        """Test that all spikes fall within specified time bounds"""
        mean_freq = np.array([25.0, 10.0])
        std_freq = np.array([5.0, 2.0])
        time_ranges = (np.array([1.5, 5.0]), np.array([3.5, 8.0]))

        spikes = self.input.generate_log_normal_spikes(
            mean_freq, std_freq, time_ranges, self.rng
        )

        # All spikes should be within the union of time ranges
        valid_range1 = (spikes >= 1.5) & (spikes < 3.5)
        valid_range2 = (spikes >= 5.0) & (spikes < 8.0)
        all_valid = valid_range1 | valid_range2

        self.assertTrue(np.all(all_valid), "All spikes should be within specified time ranges")


if __name__ == '__main__':
    unittest.main()



import setuptools, os

def get_version():
    version = {}
    with open(os.path.join("snudda", "__init__.py")) as f:
        for line in f:
            if line.startswith("__version__"):
                exec(line, version)
                break
    return version["__version__"]

if __name__ == "__main__":
    setuptools.setup(version=get_version())



import os
import sys
import numpy as np
import h5py
import pandas
import timeit


class FilterFlagserData:

    def __init__(self, flagser_data_file_name, network_meta_file_name, filtered_file_name=None):

        self.data_file_name = flagser_data_file_name
        self.network_meta_file_name = network_meta_file_name

        self.data_file = None
        self.meta_data = None
        self.filtered_data_file = None

        if filtered_file_name:
            self.filtered_file_name = filtered_file_name
        else:
            self.filtered_file_name = f"{flagser_data_file_name}-filtered.hdf5"

    def close(self):
        if self.data_file:
            self.data_file.close()
            self.data_file = None

        if self.filtered_data_file:
            self.filtered_data_file.close()
            self.filtered_data_file = None

    def load_data(self):

        self.data_file = open(self.data_file_name, "r")
        # self.meta_data = np.genfromtxt(self.network_meta_file_name, delimiter=",")
        self.meta_data = pandas.read_csv(self.network_meta_file_name,
                                         names=["neuron_id", "neuron_type", "neuron_name", "x", "y", "z", "morphology", "population_id"])

    def find_core_neurons(self, population_fraction=None, distance_to_centre=None):

        neuron_id = self.meta_data["neuron_id"].to_numpy()
        positions = self.meta_data[["x", "y", "z"]].to_numpy()

        centre_point = np.mean(positions, axis=0)
        dist_to_centre = np.linalg.norm(positions - centre_point, axis=1)

        if distance_to_centre:
            print(f"Core filtering, maximal distance to centre {distance_to_centre} meters")
            distance_threshold = distance_to_centre
        else:
            assert population_fraction is not None, f"Either distance_to_centre or population_fraction must be set"

            print(f"Filtering simplices with at least one member in core "
                  f"(core fraction {population_fraction} of all neurons)")
            distance_threshold = np.percentile(dist_to_centre, population_fraction)

        centre_idx = np.where(dist_to_centre <= distance_threshold)[0]

        lookup_table = np.zeros((len(neuron_id),), dtype=bool)
        lookup_table[centre_idx] = True

        self.filtered_data_file = open(self.filtered_file_name, "w")

        start_time = timeit.default_timer()

        keep_count = 0
        total_count = 0

        self.filtered_data_file.write("# Centre IDs:" + ", ".join([str(x) for x in neuron_id[centre_idx]]) + "\n")

        for row in self.data_file:

            neuron_id = np.array([int(x) for x in row.strip().split(" ")])

            keep_flag = np.sum(np.take(lookup_table, neuron_id)) > 0
            if keep_flag:
                self.filtered_data_file.write(",".join([f"{x}" for x in neuron_id]) + "\n")
                keep_count += 1

            total_count += 1

        self.close()

        print(f"Total time used {(timeit.default_timer() - start_time):0.1f} seconds.")
        print(f"Kept {keep_count} out of {total_count} rows")

    @staticmethod
    def get_dim_from_name(dim_name):
        return int(dim_name[6:])

    def check_if_exists(self, dim_name):
        return dim_name in self.filtered_data_file and \
               self.filtered_data_file["meta/num_simplices"][self.get_dim_from_name(dim_name)] >= 0

    def save_dim_data(self, dim_name, dim_data):
        print(f"Writing {dim_name} ({len(dim_data)} rows)")
        self.filtered_data_file.create_dataset(dim_name, data=dim_data)
        self.filtered_data_file["meta/num_simplices"][self.get_dim_from_name(dim_name)] = len(dim_data)


def filter_flagser_cli():

    import argparse
    parser = argparse.ArgumentParser("Filter flagser")
    parser.add_argument("flagser_file", help="Flagser file")
    parser.add_argument("meta_data_file", help="Meta data file")
    parser.add_argument("output_file", help="Output file name")
    parser.add_argument("--percentile", help="Percentile of population kept (0-100)", type=int, default=None)
    parser.add_argument("--distance", help="Maximum distance to centre for core neurons (micrometers)",
                        type=float, default=None)

    args = parser.parse_args()

    ff = FilterFlagserData(flagser_data_file_name=args.flagser_file,
                           network_meta_file_name=args.meta_data_file,
                           filtered_file_name=args.output_file)
    ff.load_data()
    ff.find_core_neurons(distance_to_centre=args.distance, population_fraction=args.percentile)


if __name__ == "__main__":
    filter_flagser_cli()


# TODO:
#
# What cells to record from
# What cells to give current injection to


# For each cell:
#  --

#   Pre synaptic cells (list)
#   List of current injections for the pre-synaptic cells
#
#   Post synaptic cells (list)
#   List of current injections
#
#
#

import json
import os
from collections import OrderedDict

import neuron
import numpy as np

from snudda.simulate import SnuddaSimulate

neuron.h.load_file("stdrun.hoc")


class PairRecording(SnuddaSimulate):
    """ Runs simulation with current injections to one or more neurons. This code can be used to simulate
        a subset of the neurons, ie the neurons receiving current injections and their post synaptic targets.

        First create your own network using Snudda init/place/detect/prune. Then use PairRecording to setup
        and run simulation.

        For example experiment configuration files, see Snudda/snudda/data/experiment-config/pair_recording directory

    """

    def __init__(self, network_path, experiment_config_file,
                 output_file=None, network_file=None,
                 snudda_data=None,
                 disable_gap_junctions=False,
                 disable_synapses=False,
                 log_file=None,
                 verbose=False):

        # Do stuff with experiment_config
        self.experiment_config_file = experiment_config_file
        self.experiment_config = self.read_experiment_config(experiment_config_file=experiment_config_file)

        if output_file is None:
            if "pair_recording_output_file" in self.experiment_config["meta"].keys():
                output_file = os.path.join(network_path, "simulation",
                                           self.experiment_config["meta"]["pair_recording_output_file"])
            else:
                output_file = os.path.join(network_path, "simulation", "output.hdf5")

        print(f"Pair recording output file: {output_file}")

        super().__init__(network_path=network_path,
                         network_file=network_file,
                         output_file=output_file,
                         disable_synapses=disable_synapses,
                         disable_gap_junctions=disable_gap_junctions,
                         snudda_data=snudda_data,
                         log_file=log_file,
                         verbose=verbose)

        self.simulate_neuron_ids = None

        self.sim_duration = None
        self.holding_i_clamp_list = dict()
        self.v_init_saved = dict()
        self.v_hold_saved = dict()

        # Variables for synapse current recording
        # self.synapse_currents = []
        self.record_from_pair = []

        self.parse_experiment_config()

    # def __del__(self):
    #     super().__del__()

    @staticmethod
    def read_experiment_config(experiment_config_file):

        """ Loads the experimental config from JSON file. """

        with open(experiment_config_file, "r") as f:
            return json.load(f, object_pairs_hook=OrderedDict)

    def parse_experiment_config(self):

        """ Parses the experimental config file, updating the simulation as needed. """

        if "neuron_subset" in self.experiment_config["meta"]:
            neuron_subset = self.experiment_config["meta"]["neuron_subset"]
        else:
            neuron_subset = "all"  # "prepost"

        self.set_neurons_to_simulate(neuron_subset)

        if "record_synaptic_current" in self.experiment_config:
            syn_cur_record = self.experiment_config["record_synaptic_current"]

            for pre_id, post_id in syn_cur_record:
                self.write_log(f"Marking synapses for recording: {pre_id} -> {post_id}")
                self.mark_synapses_for_recording(pre_neuron_id=pre_id, post_neuron_id=post_id)

        # Setup the network given in network_config
        super().setup()

        self.sim_duration = self.experiment_config["meta"]["simulation_duration"]

        if self.output_file is None:
            if "pair_recording_output_file" in self.experiment_config["meta"]:
                self.output_file = os.path.join(self.network_path, "simulation",
                                                self.experiment_config["meta"]["pair_recording_output_file"])
                self.record.set_new_output_file(self.output_file)

        # Setup v_init for each neuron_id specified
        if "v_init" in self.experiment_config["meta"]:

            if type(self.experiment_config["meta"]["v_init"]) == list:
                neuron_id, v_init = zip(*self.experiment_config["meta"]["v_init"])
            else:
                neuron_id = None
                v_init = self.experiment_config["meta"]["v_init"]

            self.set_v_init(neuron_id=neuron_id, v_init=v_init)

        if "v_hold" in self.experiment_config["meta"]:
            if type(self.experiment_config["meta"]["v_hold"]) == list:
                neuron_id, v_hold = zip(*self.experiment_config["meta"]["v_hold"])
            else:
                neuron_id = None
                v_hold = self.experiment_config["meta"]["v_hold"]

            self.set_v_hold(neuron_id=neuron_id, v_hold=v_hold)

        if "reversal_potential" in self.experiment_config["meta"]:
            for channel_name, v_rev in self.experiment_config["meta"]["reversal_potential"]:
                self.set_channel_rev(channel_name=channel_name, v_rev=v_rev)

        # Iterate over current injections
        for cur_info in self.experiment_config["current_injection"]:
            stim_neuron_id = self.to_list(cur_info["neuron_id"])
            stim_start_time = self.to_list(cur_info["start"])
            stim_end_time = self.to_list(cur_info["end"])

            if "amp_spread" in cur_info:
                amp_spread = cur_info["amp_spread"]
            else:
                amp_spread = None

            for nid in stim_neuron_id:

                if "amplitude" in cur_info:
                    stim_amplitude = self.to_list(cur_info["amplitude"])
                elif "requested_frequency" in cur_info:
                    requested_freq = self.to_list(cur_info["requested_frequency"])
                    stim_amplitude = self.get_corresponding_current_injection(neuron_id=nid,
                                                                              frequency=requested_freq)

                    if nid in self.holding_i_clamp_list:
                        _, cur = self.holding_i_clamp_list[nid]
                        self.write_log(f"Compensating for holding current, reducing {stim_amplitude} A by {cur} A")
                        stim_amplitude -= cur

                    v_holder = self.v_hold_saved[nid]
                else:
                    assert False, f"You need to specify 'amplitude' or 'requestedFrequency' for neuron_id {stim_neuron_id}"

                self.add_current_pulses(neuron_id=nid, start_times=stim_start_time,
                                        end_times=stim_end_time, amplitudes=stim_amplitude,
                                        amp_spread=amp_spread)

                if "noise_amplitude" in cur_info:

                    if "noise_dt" in cur_info:
                        noise_dt = cur_info["noise_dt"]  # dt = how often does noise current change
                    else:
                        noise_dt = None

                    self.add_noise(neuron_id=nid, noise_std=cur_info["noise_amplitude"],
                                   duration=self.sim_duration,
                                   dt=noise_dt)

        if "record" in self.experiment_config["meta"]:
            if self.experiment_config["meta"]["record"].lower() == "soma":
                self.add_volt_recording_soma()
            elif self.experiment_config["meta"]["record"].lower() == "all":
                self.add_volt_recording_all()
            else:
                assert False, (f"Unknown record option {self.experiment_config['meta']['record']}. "
                               f"Valid options are 'soma' or 'all'")
        else:
            # Add voltage recordings to neurons
            self.add_volt_recording_soma()

        self.setup_synaptic_recordings()

    def get_corresponding_current_injection(self, neuron_id, frequency):

        """ Extracts what current injection is needed to get requested frequency from if_info.json """

        if_file = os.path.join(self.network_info["neurons"][neuron_id]["neuron_path"], "if_info.json")
        parameter_key = self.network_info["neurons"][neuron_id]["parameter_key"]
        morphology_key = self.network_info["neurons"][neuron_id]["morphology_key"]

        assert os.path.isfile(if_file), f"Unable determine current to get frequency {frequency}, needs {if_file}"

        with open(if_file, "r") as f:
            if_data = json.load(f)

        current_list = if_data[parameter_key][morphology_key]["current"]
        frequency_list = if_data[parameter_key][morphology_key]["frequency"]

        assert (np.diff(current_list) > 0).all(), f"Injected currents are not increasing in file: {if_file} " \
                                                  f"for parameter_key {parameter_key}, morphology_key {morphology_key}"

        current = np.interp(frequency, frequency_list, current_list)

        if (np.array(frequency) < frequency_list[0]).any():
            self.write_log(f"WARNING: Neuron {neuron_id} requested frequency {frequency}, "
                           f"but lowest frequency in {if_file} is {frequency_list[0]}", force_print=True)

        if (np.array(frequency) > frequency_list[-1]).any():
            self.write_log(f"WARNING: Neuron {neuron_id} requested frequency {frequency}, "
                           f"but highest frequency in {if_file} is {frequency_list[-1]}", force_print=True)

        print(f'neuron_id: {neuron_id}, current:{current}')

        return current

    @staticmethod
    def to_list(val, new_list_len=1):

        """ If val is not a list, returns a list with new_list_len elements, each with value val

        Args:
            val : variable to upgrade to list (or leave unchanged if val is a list already)
            new_list_len : length of new list created (if val is not already a list)

        """

        if type(val) not in [list, np.ndarray, range]:
            val = [val] * new_list_len

        return val

    def set_v_hold(self, v_hold, neuron_id=None):

        """ This function is currently not called by default. If you choose to use it beware that
            the holding current may change the excitability of your neuron.

            Set holding voltage v_hold for neurons speciefied by neuron_id.
            If neuron_id = None (default), all neurons get v_hold set.

            Args:
                v_hold = Holding voltage (list or int) in volt (SI-units)
                neuron_id = Neuron ID of neurons affected (list, int or None)

            """

        if neuron_id is None:
            neuron_id = self.neuron_id
            neuron_id = self.to_list(neuron_id)
            v_hold = self.to_list(v_hold, new_list_len=len(neuron_id))
        else:
            neuron_id = [x for x in neuron_id]
            v_hold = [x for x in v_hold]

        assert self.sim_duration is not None, f"setup_holding_volt: Please set self.end_time, for holding current"

        # Setup vClamps to calculate what holding current will be needed
        soma_v_clamp = []
        soma_list = [(self.neurons[x].icell.soma[0], x) for x in neuron_id if x in self.neuron_id]

        missing_id = [x for x in self.neurons.keys() if x not in neuron_id and x in self.neuron_id]

        if len(missing_id) > 0:
            print(f"Warning: v_hold is not specified for neurons {missing_id}")

        for (s, nid), vi in zip(soma_list, v_hold):
            vc = neuron.h.SEClamp(s(0.5))
            vc.rs = 1e-9
            vc.amp1 = vi * 1e3
            vc.dur1 = 200
            soma_v_clamp.append((s, vc, nid))

        self.set_v_hold_helper(neuron_id=neuron_id, v_hold=v_hold)
        neuron.h.finitialize()

        neuron.h.tstop = 200
        neuron.h.run()

        # Dont override
        # self.holding_i_clamp_list = dict()

        assert self.sim_duration is not None, ("set_v_hold: self.end_time must be set before calling, "
                                               "IClamps need to know their duration.")

        # Setup iClamps
        for s, vc, nid in soma_v_clamp:
            cur = float(vc.i)
            ic = neuron.h.IClamp(s(0.5))
            ic.amp = cur
            ic.dur = 2 * self.sim_duration * 1e3
            self.holding_i_clamp_list[nid] = ic, cur * 1e-9

        # Remove vClamps
        v_clamps = None
        vc = None

        # Set voltage also
        self.set_v_hold_helper(neuron_id=neuron_id, v_hold=v_hold)

    def set_v_hold_helper(self, neuron_id, v_hold):
        for nid, v in zip(neuron_id, v_hold):
            self.set_resting_voltage(neuron_id=nid, rest_volt=v)
            self.v_hold_saved[nid] = v
            self.v_init_saved[nid] = v

    def set_v_init(self, neuron_id, v_init):
        """ Set initial voltage v_init for neurons speciefied by neuron_id.
            If neuron_id = None (default), all neurons get v_init set.

            Args:
                v_init = Initial voltage (list or int) in volt (SI-units)
                neuron_id = Neuron ID of neurons affected (list, int or None)

            """
        if neuron_id is None:
            neuron_id = self.neuron_id

        neuron_id = self.to_list(neuron_id)
        v_init = self.to_list(v_init, new_list_len=len(neuron_id))
        for nid, v in zip(neuron_id, v_init):
            self.set_resting_voltage(neuron_id=nid, rest_volt=v)
            self.v_init_saved[nid] = v

    def set_neurons_to_simulate(self, selection=None):

        """ Sets subset of neurons to simulate. If selection "prepost" neurons receiving current injection and
            their post synaptic targets are included. If selection is "all" or None, then all neurons are included.

            If you simulate with gap junctions, you need to make sure that each neuron simulated also have all
            gap junction coupled neuron simulated.

        Args:
            selection (string or list of int) : What neurons to include in simulation ("prepost", "all"),
                                                or a list of neuron_id to simulate

        """
        if type(selection) == list:
            sim_id = selection
            self.write_log(f"Simulating user selected neuron_ids: {sim_id}")

        if selection is None or selection.lower() == "all":
            # No restrictions, use all
            self.simulate_neuron_ids = None
            self.write_log("Simulating all neurons in network.")

        elif selection.lower() == "prepost":

            pre_id = set(sum([c["neuron_id"] if type(c["neuron_id"]) == list else [c["neuron_id"]]
                              for c in self.experiment_config["current_injection"]], []))
            sim_id = pre_id

            for pid in pre_id:
                found_syn = self.snudda_loader.find_synapses(pre_id=pid)[0]
                found_gj = self.snudda_loader.find_gap_junctions(neuron_id=pid)[0]

                if found_syn is not None:
                    post_id = set(found_syn[:, 1])
                    sim_id = sim_id.union(post_id)

                if found_gj is not None:
                    gj_id = set(found_gj[:, 1])
                    sim_id = sim_id.union(gj_id)

            self.simulate_neuron_ids = sorted(list(sim_id))

            self.write_log(f"Simulating neuron IDs {sim_id}")

        else:
            assert False, f"Unsupported selection {selection}, use 'all' or 'prepost'"

    def distribute_neurons(self):

        """ Overloading the normal distribute_neurons to allow only a subset of neurons to be simulated. """

        if self.simulate_neuron_ids is None:
            super().distribute_neurons()
        else:
            self.write_log("Only simulating some of the neurons.")

            # Only include the neurons that are in simulate_neurons_id
            assert len(self.simulate_neuron_ids) >= int(self.pc.nhost()), \
                (f"Do not allocate more workers ({int(self.pc.nhost())}) than there "
                 f"are neurons included ({len(self.simulate_neuron_ids)}).")

            idx = np.arange(int(self.pc.id()), len(self.simulate_neuron_ids), int(self.pc.nhost()))
            self.neuron_id = np.array(sorted([self.simulate_neuron_ids[x] for x in idx]))

            self.neuron_id_on_node = np.zeros((self.num_neurons,), dtype=bool)
            self.neuron_id_on_node[self.neuron_id] = True

            self.write_log(f"Node {self.pc.id()} processing neurons {self.neuron_id}")

            # Clear this variable, since old value is not valid
            self.neuron_nodes = None  # TODO, set it correctly!

            # We also need to remove gap junctions that belong to neurons not simulated
            keep_gj_flag = np.ones((self.gap_junctions.shape[0], ), dtype=bool)

            for idx, gj_row in enumerate(self.gap_junctions):
                if gj_row[0] not in self.simulate_neuron_ids or gj_row[1] not in self.simulate_neuron_ids:
                    keep_gj_flag[idx] = False

            self.gap_junctions = self.gap_junctions[keep_gj_flag, :]

    def run(self):

        """ Run simulation. """

        for nid, v in self.v_init_saved.items():
            self.set_resting_voltage(neuron_id=nid, rest_volt=v)

        # Run simulation
        super().run(self.sim_duration * 1e3, hold_v=None)

        # Write results to disk
        try:
            self.record.output_file = self.output_file
            self.write_output()

        except:
            import traceback
            t_str = traceback.format_exc()
            self.write_log(t_str)
            self.write_log(f"Saving failed, whoops. Entering debug mode.\n"
                           f"You might have had {self.output_file} opened elsewhere. Try closing it, then type:\n"
                           f"> self.write_output()\n"
                           f"If you are lucky, this will work and you wont loose any data.", is_error=True)
            import pdb
            pdb.set_trace()

    def setup_synaptic_recordings(self):
        for src_id, dest_id in self.record_from_pair:
            self.add_synapse_current_recording(src_id, dest_id)

    def plot_trace_overview(self, experiment_name=None):
        from snudda.plotting import PlotTraces

        if experiment_name is None:
            if "experiment_name" in self.experiment_config["meta"]:
                experiment_name = self.experiment_config["meta"]["experiment_name"]
            else:
                experiment_name = None

        pt = PlotTraces(output_file=self.output_file,
                        network_file=self.network_file,
                        experiment_name=experiment_name)

        pt.plot_traces([x for x in pt.voltage])

    def plot_traces(self, mark_current_y=-80.05e-3, trace_id=None, offset=150e-3):

        """
           Plot traces of post synaptic neuron activation, this assumes only one neuron was stimulated

           Args:
               mark_current_y (float) : Y-axis value to mark current injections at
               trace_id (list, optional) : Trace ID to show (default None, meaning all post synaptic traces)
           """

        for cur_info in self.experiment_config["current_injection"]:
            pre_id = cur_info["neuron_id"]
            cur_start = self.to_list(cur_info["start"])
            cur_end = self.to_list(cur_info["end"])
            cur_times = list(zip(cur_start, cur_end))

            skip_time = cur_start[0] / 2

            assert type(pre_id) == int, f"Plot traces assumes one pre-synaptic neuron stimulated: {pre_id}"
            if self.snudda_loader.find_synapses(pre_id=pre_id)[0] is None:
                post_id = []
            else:
                post_id = set(self.snudda_loader.find_synapses(pre_id=pre_id)[0][:, 1])

            experiment_name = self.get_experiment_name(empty_is_none=False)

            for pid in post_id:

                if trace_id is not None and pid not in trace_id:
                    print(f"Skipping trace {pid}, not in trace_id={trace_id}")
                    continue

                fig_name = f"Current-injection-{experiment_name}-pre-{pre_id}-post-{pid}.pdf"

                self.plot_trace(pre_id=pre_id, post_id=pid, fig_name=fig_name,
                                mark_current=cur_times, mark_current_y=mark_current_y,
                                skip_time=skip_time, offset=offset)

    def get_experiment_name(self, empty_is_none=True):
        if "experiment_name" in self.experiment_config["meta"]:
            experiment_name = self.experiment_config["meta"]["experiment_name"]
        elif empty_is_none:
            experiment_name = None
        else:
            experiment_name = ""

        return experiment_name

    def plot_trace(self, pre_id, post_id, offset=0, title=None, fig_name=None, skip_time=0,
                   mark_current=None, mark_current_y=None):

        from snudda.plotting import PlotTraces

        if not title:
            synapses, _ = self.snudda_loader.find_synapses(pre_id=pre_id, post_id=post_id)
            if synapses is not None:
                n_synapses = synapses.shape[0]
            else:
                n_synapses = 0

            title = f"{self.neurons[pre_id].name} ({pre_id}) -> " \
                    f"{self.neurons[post_id].name} ({post_id}) ({n_synapses} synapses)"

        experiment_name = self.get_experiment_name()

        pt = PlotTraces(output_file=self.output_file, network_file=self.network_file, experiment_name=experiment_name)
        pt.plot_traces(trace_id=post_id, offset=offset, title=title, fig_name=fig_name, skip_time=skip_time,
                       mark_current=mark_current, mark_current_y=mark_current_y)

    def plot_synaptic_currents(self, post_id):

        from snudda.plotting import PlotTraces

        experiment_name = self.get_experiment_name()
        pt = PlotTraces(output_file=self.output_file, network_file=self.network_file, experiment_name=experiment_name)
        pt.plot_synaptic_currents(post_id=post_id)

    def mark_synapses_for_recording(self, pre_neuron_id, post_neuron_id):

        """ What neuron pair synapses should we record current from? Add them to self.record_from_pair

            Args:
                pre_neuron_id : Neuron ID of presynaptic neuron
                post_neuron_id : Neuron ID of postsynaptic neuron

        """

        self.record_from_pair.append((pre_neuron_id, post_neuron_id))

    def set_channel_rev(self, channel_name, v_rev):

        print(f"Setting {channel_name} reversal potential to {v_rev * 1e3} mV")

        for syn_list in self.synapse_dict.values():
            for syn_ind in syn_list:
                syn = syn_ind[0]
                if channel_name == syn.hname().split("[")[0]:
                    syn.e = v_rev * 1e3




if __name__ == "__main__":
    import sys
    if '-python' in sys.argv:
        print("Network_simulate.py called through nrniv, fixing arguments")
        pythonidx = sys.argv.index('-python')
        if len(sys.argv) > pythonidx:
            sys.argv = sys.argv[pythonidx + 1:]

    from argparse import ArgumentParser, RawTextHelpFormatter
    parser = ArgumentParser("Pair recording", formatter_class=RawTextHelpFormatter)
    parser.add_argument("network_path")
    parser.add_argument("experiment_config_file")
    args = parser.parse_args()
    pr = PairRecording(network_path=args.network_path, experiment_config_file=args.experiment_config_file)
    pr.run()



import os

import numpy as np
from numba import jit

import matplotlib.pyplot as plt
from snudda.utils.load_network_simulation import SnuddaLoadSimulation


class PlotCrossCorrelogram:

    def __init__(self, simulation_file=None, snudda_simulation_load=None):

        if snudda_simulation_load:
            self.sim_data = snudda_simulation_load
        else:
            self.sim_data = SnuddaLoadSimulation(network_simulation_output_file=simulation_file)

    def calculate_all_pair_cross_correlogram(self, neuron_id, time_range=None, shuffle_correct=True,
                                             n_bins=101, width=50e-3):

        bin_count_total = None
        bin_edges = None
        n_traces = 0

        shuffle_count_total = None

        spike_data = self.sim_data.get_spikes(neuron_id=neuron_id)

        for na in spike_data.keys():

            if spike_data[na].size == 0:
                continue

            for nb in spike_data.keys():
                if na == nb:
                    continue

                if spike_data[nb].size == 0:
                    continue

                bin_count, edges = self.calculate_cross_correlogram(spike_data[na],
                                                                    spike_data[nb],
                                                                    time_range=time_range,
                                                                    n_bins=n_bins, width=width)
                
                shuffle_count, shuffle_edges = \
                    self.calculate_cross_correlogram(self.shuffle_spikes(spike_data[na], time_range=time_range),
                                                     self.shuffle_spikes(spike_data[nb], time_range=time_range),
                                                     time_range=time_range,
                                                     n_bins=n_bins, width=width)

                assert (edges == shuffle_edges).all()
                
                if bin_edges is None:
                    bin_edges = edges
                    bin_count_total = bin_count
                else:
                    assert (bin_edges == edges).all()
                    bin_count_total += bin_count

                if shuffle_count_total is None:
                    shuffle_count_total = shuffle_count
                else:
                    shuffle_count_total += shuffle_count

                n_traces += 1

        if shuffle_correct:
            bin_count_total -= shuffle_count_total

        fraction = bin_count_total / n_traces

        return fraction, bin_edges

    @staticmethod
    @jit(nopython=True, fastmath=True, cache=True)
    def calculate_cross_correlogram(spike_times_a, spike_times_b, n_bins=101, width=50e-3, time_range=None,
                                    empty_array=np.array([], dtype=np.float64)):

        """ This code assumes spikes are ordered. """

        assert spike_times_a.shape[0] == 1 and spike_times_b.shape[0] == 1
        
        if time_range is not None:
            idx_a_start = 0
            len_a = spike_times_a.size
            
            while idx_a_start < len_a and spike_times_a[0][idx_a_start] < time_range[0]:
                idx_a_start += 1

            idx_a_end = idx_a_start
            while idx_a_end < len_a and spike_times_a[0][idx_a_end] < time_range[1]:
                idx_a_end += 1

            idx_a = np.arange(idx_a_start, idx_a_end)

            idx_b_start = 0
            len_b = spike_times_b.size
            while idx_b_start < len_b and spike_times_b[0][idx_b_start] < time_range[0]:
                idx_b_start += 1

            idx_b_end = idx_b_start
            while idx_b_end < len_b and spike_times_b[0][idx_b_end] < time_range[1]:
                idx_b_end += 1

            idx_b = np.arange(idx_b_start, idx_b_end)

            # idx_a2 = np.where(np.logical_and(time_range[0] <= spike_times_a,
            #                                  spike_times_a <= time_range[1]))[1]

            # idx_b2 = np.where(np.logical_and(time_range[0] <= spike_times_b,
            #                                  spike_times_b <= time_range[1]))[1]

            if len(idx_a) == 0 or len(idx_b) == 0:
                t_diff = empty_array
            else:
                t_diff = (np.kron(spike_times_a[:, idx_a], np.ones(spike_times_b[:, idx_b].shape).T)
                          - np.kron(np.ones(spike_times_a[:, idx_a].shape), spike_times_b[:, idx_b].T)).flatten()

        else:

            t_diff = (np.kron(spike_times_a, np.ones(spike_times_b.shape).T)
                      - np.kron(np.ones(spike_times_a.shape), spike_times_b.T)).flatten()

        # t_diff = t_diff[np.where(np.abs(t_diff) <= width)[0]]

        bin_count, bin_edges = np.histogram(t_diff, bins=n_bins, range=[-width, width])
        return bin_count, bin_edges    

    def get_range(self, spike_times, time_range):

        idx = np.where(np.logical_and(time_range[0] <= spike_times.flatten(),
                                      spike_times.flatten() <= time_range[1]))[0]

        return spike_times[:, idx]
    
    def shuffle_spikes(self, spike_times, time_range):

        if time_range:
            
            idx = np.where(np.logical_and(time_range[0] <= spike_times.flatten(),
                                          spike_times.flatten() <= time_range[1]))[0]

            if np.size(idx) == 0:
                return np.array([[]])
            
            isi = np.diff(spike_times.flatten()[idx])
            isi = np.append(isi, spike_times.flatten()[idx[0]] - time_range[0])
            st = np.cumsum(np.random.permutation(isi)).reshape(1, (len(idx)))
            st += np.random.uniform(low=time_range[0], high=time_range[0])
            st = np.sort(st % (time_range[1] - time_range[0]) + time_range[0])
            
            return st
            
        isi = np.diff(spike_times)
        isi = np.append(isi, spike_times.flatten()[0])
        
        return np.cumsum(np.random.permutation(isi)).reshape(spike_times.shape)

    def plot_cross_correlogram(self, spike_times_a, spike_times_b, fig_file_name=None):

        bin_count, bin_edges = self.calculate_cross_correlogram(spike_times_a=spike_times_a,
                                                                spike_times_b=spike_times_b)
        plt.rcParams.update({'font.size': 24,
                             'xtick.labelsize': 20,
                             'ytick.labelsize': 20,
                             'legend.loc': 'best'})

        plt.figure()
        plt.stairs(values=bin_count, edges=bin_edges)
        plt.xlabel("Time (s)", fontsize=20)
        plt.ylabel("Count", fontsize=20)
        plt.ion()
        plt.show()

        if fig_file_name:
            plt.savefig(fig_file_name, dpi=300)

    def plot_all_pair_cross_correlogram(self, neuron_id, fig_file_name=None, time_range=None, shuffle_correct=True,
                                        ax=None, linestyle="-", label=None, colour="black", fig_size=None,
                                        legend_loc="best", show_figure=True, n_bins=101, width=50e-3):

        fraction, bin_edges = self.calculate_all_pair_cross_correlogram(neuron_id=neuron_id, time_range=time_range,
                                                                        shuffle_correct=shuffle_correct,
                                                                        n_bins=n_bins, width=width)

        plt.rcParams.update({'font.size': 24,
                             'xtick.labelsize': 20,
                             'ytick.labelsize': 20,
                             'legend.loc': legend_loc})

        if fraction is not None:

            if ax is None:
                fig = plt.figure(figsize=fig_size)
                ax = fig.add_subplot()

            ax.stairs(values=fraction, edges=bin_edges, linestyle=linestyle, label=label, color=colour)
            ax.legend()
            plt.xlabel("Time (s)", fontsize=20)
            plt.ylabel("Normalised count", fontsize=20)
            if time_range is not None:
                ax.title(f"From {time_range[0]}s to {time_range[1]}s")

            if show_figure:
                plt.ion()
                plt.show()

            if fig_file_name:
                if not os.path.isdir(os.path.dirname(fig_file_name)):
                    print(f"Creating directory {os.path.dirname(fig_file_name)}")
                    os.mkdir(os.path.dirname(fig_file_name))

                plt.savefig(fig_file_name, dpi=300)
        else:
            print(f"No spikes, skipping spike histogram {fig_file_name}")

        return ax


def cleanup_memory(d_view, variable_list):
    """ Clear variables in variable_list on workers specified by d_view.

    Args:
        d_view : ipyparallel direct view
        variable_list : list of variables to clear
        """

    for var in variable_list:
        d_view.execute(f"if '{var}' in locals(): del {var}")
        # d_view.execute(f"{var}=None")


# Cleans up memory on workers after execution
def cleanup(rc, state):
    """ When place, detect, prune are run one after the other, we do not need to keep them all in memory
    at the same time. As one finishes, we can clear the memory.

    Args:
        rc : ipyparallel remote client
        state : State to free memory (e.g. "place", "detect", "project", "prune"
        """

    if rc is not None:

        # Check if the client is still active
        try:
            if hasattr(rc, '_closed') and rc._closed:
                return
            # Alternative check: try to access a property that would fail if closed
            _ = rc.ids
        except (RuntimeError, AttributeError):
            # Client is closed or invalid, skip cleanup
            return

        d_view = rc.direct_view(targets='all')

        var_lookup = {"place": ["inner_mask", "sm"],  # region_mesh_vedo.py
                      "detect": ["min_max", "result", "sd"],
                      "project": [],  # Currently does not support parallel execution
                      "prune": ["syn_before", "syn_after", "merge_result_syn", "merge_result_gj", "sp"]}

        if state in var_lookup:
            cleanup_memory(d_view, var_lookup[state])
        else:
            print(f"cleanup: Unknown state: {state}")



import numpy as np
import json
import os
import scipy
import scipy.optimize

import pyswarms as ps
from run_little_synapse_run import RunLittleSynapseRun
import matplotlib.pyplot as plt
import matplotlib

from snudda.utils.numpy_encoder import NumpyEncoder

############################################################################


class Planert2010part2(object):

    def __init__(self, data_type, cell_id=None, pretty_plot=True, fit_data=True):

        self.data_legend = {"II": "iSPN to iSPN",
                            "ID": "iSPN to dSPN",
                            "DD": "dSPN to dSPN",
                            "DI": "dSPN to iSPN",
                            "FD": "FSN to dSPN",
                            "FI": "FSN to iSPN"}

        # Since we reuse code that was parallel, need to say that this is master
        self.role = "master"
        self.fig_resolution = 300
        self.pretty_plot = pretty_plot
        self.parameter_cache = dict([])
        self.rsr_synapse_model = None

        file_name = os.path.join("DATA", "Planert2010", f"PlanertFitting-{data_type}-cache.json")
        print(f"Loading data {data_type}: {file_name}")

        with open(file_name, "r") as f:
            self.data = json.load(f)

        if cell_id is None:
            cell_id = np.arange(0, len(self.data["simAmp"]))

        # Also load old parameters if they exist?
        self.cache_file_name = os.path.join("DATA", "Planert2010", f"PlanertFitting-{data_type}-tmgaba-fit.json")

        self.load_parameter_cache()

        if fit_data:
            # Save parameters after fitting

            if type(cell_id) in [list, np.ndarray]:
                for cID in cell_id:
                    self.fit_peaks(data_type, cID)
            else:
                self.fit_peaks(data_type, int(cell_id))

            self.save_parameter_cache()
        else:
            self.setup_model_synapse_fitting(data_type)

            if type(cell_id) in [list, np.ndarray]:
                for cID in cell_id:
                    self.plot_data(data_type, cID, run_sim=True, show=False)
            else:
                self.plot_data(data_type, int(cell_id), run_sim=True, show=False)

    ############################################################################

    def __delete__(self):
        # Save cache file before exiting
        self.save_parameter_cache()

    def add_parameter_cache(self, cell_id, name, value):

        if cell_id not in self.parameter_cache:
            self.parameter_cache[int(cell_id)] = dict([])

        self.parameter_cache[int(cell_id)][name] = value

    ############################################################################

    def get_parameter_cache(self, cell_id, name):

        if cell_id in self.parameter_cache and name in self.parameter_cache[cell_id]:
            return self.parameter_cache[cell_id][name]
        else:
            return None

    ############################################################################

    def load_parameter_cache(self):

        if os.path.exists(self.cache_file_name):
            try:
                print(f"Loading cache file {self.cache_file_name}")
                with open(self.cache_file_name, "r") as f:
                    tmp_dict = json.load(f)

                    self.parameter_cache = dict([])
                    for k in tmp_dict:
                        self.parameter_cache[int(k)] = tmp_dict[k]

                    f.close()
            except:
                import traceback
                tstr = traceback.format_exc()
                print(tstr)

                print(f"Unable to open {self.cache_file_name}")
                self.parameter_cache = dict([])
        else:
            # No cache file to load, create empty dictionary
            self.parameter_cache = dict([])

            ############################################################################

    def save_parameter_cache(self):

        if self.role != "master":
            print("No servants are allowed to write output to json, ignoring call.")
            return

        print("Saving parameters to cache file: " + str(self.cache_file_name))

        try:
            with open(self.cache_file_name, "w") as f:
                json.dump(self.parameter_cache, f, indent=2, cls=NumpyEncoder)
                f.close()
        except:
            import traceback
            tstr = traceback.format_exc()
            print(tstr)

            print(f"Failed to save cache file ... {self.cache_file_name}")

    ############################################################################

    def fit_peaks(self, data_type, cell_id):

        print(f"Optimising for ID = {cell_id}")

        # Get peaks
        peak_height = np.array(self.data["simAmp"][cell_id])
        t_stim = np.array(self.data["tStim"])

        assert len(t_stim) == len(peak_height), "Inconsistent data lengths for fitPeaks"
        sigma = np.ones(len(peak_height))

        # Setup neuron model
        self.setup_model_synapse_fitting(data_type)

        # U, tauR, tauF, tauRatio, cond (obs, tau = tauRatio * tauR)
        model_bounds = ([1e-3, 1e-4, 1e-4, 0, 1e-5],
                        [1.0, 2, 2, 0.9999999, 1e-1])

        # Pyswarm options
        options = {"c1": 0.5, "c2": 0.3, "w": 0.9}  # default

        n_particles = 200
        n_iterations = 10

        optimizer = ps.single.GlobalBestPSO(n_particles=n_particles,
                                            dimensions=len(model_bounds[0]),
                                            options=options,
                                            bounds=model_bounds)

        cost, fit_params = optimizer.optimize(self.neuron_synapse_swarm_helper_planert,
                                              iters=n_iterations,
                                              t_stim=t_stim,
                                              peak_height=peak_height)

        model_heights, t_sim, v_sim = self._neuron_synapse_swarm_helper(fit_params, t_stim)

        # tau < tauR, so we use tauRatio for optimisation
        fit_params[3] *= fit_params[1]  # tau = tauR * tauRatio

        print("Parameters: U = %.3g, tauR = %.3g, tauF = %.3g, tau = %.3g, cond = %3.g" % tuple(fit_params))

        print(f"peakHeight = {peak_height}")
        print(f"modelHeights = {model_heights}")

        self.add_parameter_cache(cell_id, "synapse",
                                 {"U": fit_params[0],
                                  "tauR": fit_params[1],
                                  "tauF": fit_params[2],
                                  "tau": fit_params[3],
                                  "cond": fit_params[4]})

        self.add_parameter_cache(cell_id, "surrogatePeaks", peak_height)
        self.add_parameter_cache(cell_id, "fittedPeaks", model_heights)

        self.plot_data(data_type=data_type, cell_id=cell_id,
                       t_stim=t_stim,
                       surrogate_peaks=peak_height,
                       model_peaks=model_heights,
                       t_sim=t_sim,
                       v_sim=v_sim,
                       model_params=fit_params)

    ############################################################################

    def plot_data(self, data_type, cell_id,
                  t_stim=None,
                  surrogate_peaks=None,
                  model_peaks=None,
                  t_sim=None, v_sim=None,
                  model_params=None,
                  show=True,
                  t_skip=0.01,
                  pretty_plot=None,
                  run_sim=False):

        matplotlib.rcParams.update({'font.size': 24})

        if surrogate_peaks is None:
            surrogate_peaks = np.array(self.data["simAmp"][cell_id])

        if t_stim is None:
            t_stim = np.array(self.data["tStim"])

        if model_params is None:
            p_dict = self.get_parameter_cache(cell_id, "synapse")
            model_params = [p_dict[x] for x in ["U", "tauR", "tauF", "tau", "cond"]]

        if run_sim:

            assert model_params is not None, \
                "plot_data: model_params must be given if run_sim = True"
            if t_sim is not None or v_sim is not None:
                print("Ignoring t_sim and v_sim when run_sim = True")

            print(f"Run simulation for {data_type} {cell_id}")
            U, tau_r, tau_f, tau, cond = model_params

            model_peaks, t_sim, v_sim = self.synapse_model_neuron(t_stim, U,
                                                                  tau_r, tau_f, cond, tau,
                                                                  params={},
                                                                  return_trace=True)

        if pretty_plot is None:
            pretty_plot = self.pretty_plot

        plt.figure()
        if t_sim is not None:

            t_idx = np.where(t_sim > t_skip)[0][0]

            v_base = np.mean(v_sim[-50:-1] * 1e3)

            for t, sp, mp in zip(t_stim * 1e3, surrogate_peaks * 1e3, model_peaks * 1e3):
                idx = np.argmin(np.abs(t_sim - t))
                plt.plot([t, t], [v_base, v_base + sp], color=(1, 0.31, 0), linewidth=3)
                # plt.plot([t,t],[vBase,vBase+sp],color="red",linewidth=3)
                plt.plot([t, t], [v_base, v_base + mp], color="blue")

            plt.plot(1e3 * t_sim[t_idx:], 1e3 * v_sim[t_idx:], color="black")

        else:
            for t, sp, mp in zip(t_stim * 1e3, surrogate_peaks * 1e3, model_peaks * 1e3):
                # plt.plot([t,t],[0,sp],color="red",linewidth=3)
                plt.plot([t, t], [0, sp], color=(1, 0.31, 0), linewidth=3)
                plt.plot([t, t], [0, mp], color="blue")

            v_base = 0  # No sim data, base it at 0

        if pretty_plot:

            v_bar_length = 0.05
            if data_type[0] == "F":
                v_bar_length = 0.5

            # Draw scalebars
            v_scale_x = 1200
            # vMax = np.max(vPlot[np.where(tPlot > 0.050)[0]])
            y_scale_bar = v_base + float(np.diff(plt.ylim())) / 4
            v_scale_y1 = y_scale_bar + v_bar_length
            v_scale_y2 = y_scale_bar
            t_scale_y = y_scale_bar
            t_scale_x1 = v_scale_x
            t_scale_x2 = v_scale_x + 100

            plt.plot([v_scale_x, v_scale_x], [v_scale_y1, v_scale_y2], color="black")
            plt.plot([t_scale_x1, t_scale_x2], [t_scale_y, t_scale_y], color="black")

            plt.text(v_scale_x - 100, v_scale_y2 + 0.30 * float(np.diff(plt.ylim())),
                     ("%.2f" % (v_scale_y1 - v_scale_y2)) + " mV",
                     rotation=90)
            plt.text(v_scale_x, v_scale_y2 - float(np.diff(plt.ylim())) / 10,
                     ("%.0f" % (t_scale_x2 - t_scale_x1) + " ms"))

            plt.axis("off")

            if False:
                plt.ion()
                plt.show()
                import pdb
                pdb.set_trace()

        if not pretty_plot:
            if model_params is not None:
                title_str = "\nU=%.3g, tauR=%.3g, tauF=%.3g, tau=%.3g,\ncond=%.3g" \
                           % (model_params[0],
                              model_params[1],
                              model_params[2],
                              model_params[3],
                              model_params[4])
                plt.title(title_str)

            plt.xlabel("Time (ms)")
            plt.ylabel("Volt (mV)")

            # Remove part of the frame
            plt.gca().spines["right"].set_visible(False)
            plt.gca().spines["top"].set_visible(False)

        else:
            if data_type in self.data_legend:
                plt.title(self.data_legend[data_type])

            plt.axis("off")

        if not os.path.exists("figures/"):
            os.makedirs("figures/")

        if pretty_plot:
            fig_name = os.path.join("figures", f"PlanertFitting-{data_type}-{cell_id}-noaxis.pdf")
        else:
            fig_name = os.path.join("figures", f"PlanertFitting-{data_type}-{cell_id}.pdf")

        plt.savefig(fig_name, dpi=self.fig_resolution)

        if show:
            plt.ion()
            plt.show()
        else:
            plt.ioff()
            plt.close()

    ############################################################################

    def neuron_synapse_swarm_helper_planert(self, pars, t_stim, peak_height):

        res = np.zeros((pars.shape[0]))

        for idx, p in enumerate(pars):
            peak_h, t_sim, v_sim = self._neuron_synapse_swarm_helper(p, t_stim)

            # Calculating error in peak height
            h_diff = np.abs(peak_h - peak_height)
            h_diff[0] *= 3
            h_diff[-1] *= 3
            h_error = np.sum(h_diff) / len(h_diff)

            res[idx] = h_error

        return res

    ############################################################################

    def _neuron_synapse_swarm_helper(self,
                                     pars,
                                     t_spikes):

        U, tau_r, tau_f, tau_ratio, cond = pars
        tau = tau_r * tau_ratio
        params = {}

        peak_heights, t_sim, v_sim = self.synapse_model_neuron(t_spikes, U,
                                                               tau_r, tau_f, cond, tau,
                                                               params=params,
                                                               return_trace=True)

        return peak_heights, t_sim, v_sim

    ############################################################################

    def setup_model_synapse_fitting(self, data_type, params=None):

        # If the delta model existed, clear it
        # self.rsrDeltaModel = None

        soma_diameter = 20e-6
        soma_gleak = 3

        if not params:
            params = {"somaDiameter": soma_diameter, "somaGleak": soma_gleak}

        t_stim = np.array(self.data["tStim"])
        max_time = np.max(t_stim) + 0.5

        baseline_depol = -80e-3

        self.rsr_synapse_model = RunLittleSynapseRun(stim_times=t_stim,
                                                     holding_voltage=baseline_depol,
                                                     synapse_type="GABA",
                                                     params=params,
                                                     time=max_time)

    ############################################################################

    def synapse_model_neuron(self, t_spike, U, tau_r, tau_f, cond, tau,
                             params=None,
                             return_trace=False):

        # print("Running neuron model")

        assert self.rsr_synapse_model is not None, \
            "!!! Need to call setupModelSynapseFitting first"

        if not params:
            params = {}

        # Should we make a copy of params, to not destroy it? ;)
        params["U"] = U
        params["tauR"] = tau_r
        params["tauF"] = tau_f
        params["cond"] = cond
        params["tau"] = tau

        # print("params=" + str(params))

        (tSim, vSim, iSim) = \
            self.rsr_synapse_model.run2(pars=params)

        if tSim.shape != vSim.shape:
            print("Shape are different, why?!")
            import pdb
            pdb.set_trace()

        peak_idx = self.get_peak_idx2(time=tSim, volt=vSim, stim_time=t_spike)
        peakHeight, decayFits, vBase = self.find_trace_heights(tSim, vSim, peak_idx)

        if return_trace:
            return peakHeight, tSim, vSim
        else:
            return peakHeight

    ############################################################################

    def get_peak_idx2(self, stim_time, time, volt):

        freq = 1.0 / (stim_time[1] - stim_time[0])
        p_window = 1.0 / (2 * freq) * np.ones(stim_time.shape)
        p_window[-1] *= 5

        peak_info = self.find_peaks_helper(p_time=stim_time,
                                           p_window=p_window,
                                           time=time,
                                           volt=volt)

        return peak_info["peakIdx"]

    ############################################################################

    # Find peaks within pStart[i] and pStart[i]+pWindow[i]
    # The value is not the amplitude of the peak, just the voltage at the peak

    @staticmethod
    def find_peaks_helper(p_time, p_window, time=None, volt=None):

        peak_idx = []
        peak_time = []
        peak_volt = []

        for pt, pw in zip(p_time, p_window):
            t_start = pt
            t_end = pt + pw

            t_idx = np.where(np.logical_and(t_start <= time, time <= t_end))[0]

            # We assume that neuron is more depolarised than -65, ie gaba is
            # also depolarising
            p_idx = t_idx[np.argmax(volt[t_idx])]

            peak_idx.append(int(p_idx))
            peak_time.append(time[p_idx])
            peak_volt.append(volt[p_idx])

        # Save to cache -- obs peakVolt is NOT amplitude of peak, just volt

        peak_dict = {"peakIdx": np.array(peak_idx),
                     "peakTime": np.array(peak_time),
                     "peakVolt": np.array(peak_volt)}  # NOT AMPLITUDE

        return peak_dict

    # (peakIdx,peakTime,peakVolt)

    ############################################################################

    # We are using a simplified function that skips decay fits

    @staticmethod
    def find_trace_heights(time, volt, peak_idx):

        decay_func = lambda x, a, b, c: a * np.exp(-x / b) + c

        v_base = np.mean(volt[int(0.3 * peak_idx[0]):int(0.8 * peak_idx[0])])

        peak_height = np.zeros((len(peak_idx, )))
        peak_height[0] = volt[peak_idx[0]] - v_base

        decay_fits = []

        for idx_b in range(1, len(peak_idx)):

            if peak_height[0] > 0:
                if idx_b < len(peak_idx) - 1:
                    p0d = [0.06, -0.05, -0.074]
                else:
                    p0d = [1e-8, 10000, -0.0798]
            else:
                # In some cases for GABA we had really fast decay back
                if idx_b < len(peak_idx) - 1:
                    p0d = [-0.06, 0.05, -0.0798]
                else:
                    p0d = [-1e-5, 1e5, -0.0798]

            idx_a = idx_b - 1

            peak_idx_a = peak_idx[idx_b - 1]  # Prior peak
            peak_idx_b = peak_idx[idx_b]  # Next peak

            if idx_b < len(peak_idx) - 1:
                # Not the last spike

                idx_start = int(peak_idx_a * 0.2 + peak_idx_b * 0.8)
                idx_end = int(peak_idx_a * 0.08 + peak_idx_b * 0.92)
            else:
                # Last spike, use only last half of decay trace
                idx_start = int(peak_idx_a * 0.6 + peak_idx_b * 0.4)
                idx_end = int(peak_idx_a * 0.08 + peak_idx_b * 0.92)

            try:
                assert idx_start < idx_end
            except:
                import traceback
                t_str = traceback.format_exc()
                print(t_str)

                plt.figure()
                plt.plot(time, volt)
                plt.xlabel("Time (error plot)")
                plt.ylabel("Volt (error plot)")
                plt.ion()
                plt.show()
                plt.title("ERROR!!!")
                import pdb
                pdb.set_trace()

            t_ab = time[idx_start:idx_end]
            v_ab = volt[idx_start:idx_end]

            t_ab_fit = t_ab - t_ab[0]
            v_ab_fit = v_ab

            try:

                try:
                    fit_params, p_cov = scipy.optimize.curve_fit(decay_func, t_ab_fit, v_ab_fit, p0=p0d)
                except:
                    import traceback
                    t_str = traceback.format_exc()
                    print(t_str)

                    print("!!! Failed to converge, trying with smaller decay constant")
                    p0d[1] *= 0.01
                    fit_params, p_cov = scipy.optimize.curve_fit(decay_func, t_ab_fit, v_ab_fit,
                                                                 p0=p0d)

                t_b = time[peak_idx_b] - t_ab[0]
                v_base_b = decay_func(t_b, fit_params[0], fit_params[1], fit_params[2])

                peak_height[idx_b] = volt[peak_idx_b] - v_base_b

                v_fit = decay_func(t_ab - t_ab[0], fit_params[0], fit_params[1], fit_params[2])
                decay_fits.append((t_ab, v_fit))

                ################################################################

                if False:
                    plt.figure()
                    plt.plot(t_ab, v_ab, 'r')
                    plt.title("Error in findTraceHeights")
                    plt.xlabel("time")
                    plt.ylabel("volt")
                    plt.plot(t_ab, v_fit, 'k-')
                    plt.ion()
                    plt.show()

                    import pdb
                    pdb.set_trace()

                ########################################

            except:

                print("Check that the threshold in the peak detection before is OK")
                # self.plot(name)
                import traceback
                t_str = traceback.format_exc()
                print(t_str)

                if True:
                    plt.figure()
                    plt.plot(t_ab, v_ab, 'r')
                    plt.title("Error in findTraceHeights")
                    plt.xlabel("time")
                    plt.ylabel("volt")
                    # plt.plot(tAB,vFit,'k-')
                    plt.ion()
                    plt.show()

                import pdb
                pdb.set_trace()

        # import pdb
        # pdb.set_trace()

        return peak_height.copy(), decay_fits, v_base

    ############################################################################

    ############################################################################


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(description="Fit GABA model to peaks")
    parser.add_argument("dataType", choices=["DD", "ID", "DI", "II", "FI", "FD"])
    parser.add_argument("--plotOnly", help="Only plot, no new fitting",
                        action="store_true")
    args = parser.parse_args()

    if args.plotOnly:
        fitData = False
    else:
        fitData = True

    pp = Planert2010part2(data_type=args.dataType, cell_id=None,
                          pretty_plot=True, fit_data=fitData)
    # pp = Planert2010part2(dataType=args.dataType,cellID=0)



from argparse import ArgumentParser, RawTextHelpFormatter
from snudda.core import Snudda
from snudda.help import snudda_help_text
from snudda.utils.benchmark_logging import BenchmarkLogging
import os
import sys


def snudda_cli():
    """
    Command line parser for Snudda.
    Valid actions are: init, place, detect, prune, input, simulate, help
    """

    if '-python' in sys.argv:
        print("Snudda's cli.py called through nrniv, fixing arguments")
        pythonidx = sys.argv.index('-python')
        if len(sys.argv) > pythonidx:
            sys.argv = sys.argv[pythonidx + 1:]

    # print(f"Current working directory: {os.getcwd()}")
    
    parser = ArgumentParser(description=f"Snudda microcircuit generator\n\n{snudda_help_text()}",
                            formatter_class=RawTextHelpFormatter)

    sub_parsers = parser.add_subparsers(help="action", dest="action")
    sub_parsers.required = True

    create_parser = sub_parsers.add_parser("create")
    create_parser.add_argument("path", help="Location of network")
    create_parser.add_argument("-input", "--input", help="Input config")
    create_parser.add_argument("-overwrite", "--overwrite", help="Allow overwriting of old directory", action="store_true")
    create_parser.add_argument("-parallel", "--parallel", action="store_true", default=False)
    create_parser.add_argument("--profile", help="Run python cProfile", action="store_true")
    create_parser.add_argument("--verbose", action="store_true")

    init_parser = sub_parsers.add_parser("init")
    init_parser.add_argument("path", help="Location of network")
    # init_parser.add_argument("size", type=int, help="Number of neurons in network", default=None)
    init_parser.add_argument("-size", "--size", dest="size",
                             type=int, help="Number of neurons in network", default=None)
    init_parser.add_argument("--snudda_data", "--SnuddaData", type=str, default=None, dest="snudda_data",
                             help="Path to SNUDDA_DATA")
    init_parser.add_argument("--neurons_dir", type=str, default=None,
                             help="Path to neurons_dir, default is $DATA/neurons (DEPRECATED, use --snudda_data instead")
    init_parser.add_argument("-overwrite", "--overwrite", help="Allow overwriting of old directory",
                             action="store_true")
    init_parser.add_argument("--connection_file", "-connectionFile", "--connectionFile", default=None, dest="connection_file",
                             help="Use connectivity from user specified JSON file")
    init_parser.add_argument("--stay_inside", "--honorStayInside", "--stayInside", default=False, dest="stay_inside", action="store_true")
    init_parser.add_argument("-randomseed", "--randomseed", "--seed", default=None, help="Random seed", type=int)
    init_parser.add_argument("--profile", help="Run python cProfile", action="store_true")
    init_parser.add_argument("--verbose", action="store_true")

    import_parser = sub_parsers.add_parser("import")
    import_parser.add_argument("path", help="Location of network")
    import_parser.add_argument("config_file", help="Location of config_file to import")
    import_parser.add_argument("--snudda_data", help="Location of snudda_data", default=None)
    import_parser.add_argument("-overwrite", "--overwrite", action="store_true", help="Overwrite old config file")
    import_parser.add_argument("--profile", help="Run python cProfile", action="store_true")
    import_parser.add_argument("--verbose", action="store_true")

    place_parser = sub_parsers.add_parser("place")
    place_parser.add_argument("path", help="Location of network")
    place_parser.add_argument("-randomseed", "--randomseed", "--seed", default=None, help="Random seed", type=int)
    place_parser.add_argument("--stay_inside", "--honorStayInside", "--stayInside", dest="stay_inside", default=False, action="store_true")
    place_parser.add_argument("--profile", help="Run python cProfile", action="store_true")
    place_parser.add_argument("--verbose", action="store_true")
    place_parser.add_argument("--h5legacy", help="Use legacy hdf5 support", action="store_true")
    place_parser.add_argument("-parallel", "--parallel", action="store_true", default=False)
    place_parser.add_argument("-ipython_profile", "--ipython_profile", default=None)
    place_parser.add_argument("-ipython_timeout", "--ipython_timeout", default=120, type=int)

    detect_parser = sub_parsers.add_parser("detect")
    detect_parser.add_argument("path", help="Location of network")
    detect_parser.add_argument("-randomseed", "--randomseed", "--seed", default=None, help="Random seed", type=int)
    detect_parser.add_argument("-cont", "--cont", help="Continue partial touch detection", action="store_true")
    detect_parser.add_argument("-hvsize", "--hvsize", default=100,
                               help="Hyper voxel size, eg. 100 = 100x100x100 voxels in hypervoxel")
    detect_parser.add_argument("--volume_id", "--volumeID", help="Specify volume ID for detection step")
    detect_parser.add_argument("--profile", help="Run python cProfile", action="store_true")
    detect_parser.add_argument("--verbose", action="store_true")
    detect_parser.add_argument("--h5legacy", help="Use legacy hdf5 support", action="store_true")
    detect_parser.add_argument("-parallel", "--parallel", action="store_true", default=False)
    detect_parser.add_argument("-ipython_profile", "--ipython_profile", default=None)
    detect_parser.add_argument("-ipython_timeout", "--ipython_timeout", default=120, type=int)

    prune_parser = sub_parsers.add_parser("prune")
    prune_parser.add_argument("path", help="Location of network")
    prune_parser.add_argument("-randomseed", "--randomseed", "--seed", default=None, help="Random seed", type=int)
    prune_parser.add_argument("--config_file", "--configFile", dest="config_file", default=None,
                              help="Prune using different network config file, useful when tuning pruning")
    prune_parser.add_argument("--profile", help="Run python cProfile", action="store_true")
    prune_parser.add_argument("--verbose", action="store_true")
    prune_parser.add_argument("--h5legacy", help="Use legacy hdf5 support", action="store_true")
    prune_parser.add_argument("--keepfiles", action="store_true",
                              help="Keep temp and voxel files after pruning (e.g. useful if you want to rerun pruning)")
    prune_parser.add_argument("--save_putative", "--savePutative", action="store_true",
                              help="Also saved network-putative-synapses.hdf5 with unpruned network")
    prune_parser.add_argument("-parallel", "--parallel", action="store_true", default=False)
    prune_parser.add_argument("-ipython_profile", "--ipython_profile", default=None)
    prune_parser.add_argument("-ipython_timeout", "--ipython_timeout", default=120, type=int)

    input_parser = sub_parsers.add_parser("input")
    input_parser.add_argument("path", help="Location of network")
    input_parser.add_argument("--input", help="Input json config file (for input setup)", default=None)
    input_parser.add_argument("--inputFile", help="Input hdf5 file (for simulation)",
                              dest="input_file", default=None)
    input_parser.add_argument("--networkFile", help="Network file, if not network-synapses.hdf5",
                              dest="network_file")
    input_parser.add_argument("--time", type=float, default=None, help="Duration of simulation in seconds")
    input_parser.add_argument("-randomseed", "--randomseed", "--seed", default=None, help="Random seed", type=int)
    input_parser.add_argument("--profile", help="Run python cProfile", action="store_true")
    input_parser.add_argument("--verbose", action="store_true")
    input_parser.add_argument("--h5legacy", help="Use legacy hdf5 support", action="store_true")
    input_parser.add_argument("-parallel", "--parallel", action="store_true", default=False)
    input_parser.add_argument("-ipython_profile", "--ipython_profile", default=None)
    input_parser.add_argument("-ipython_timeout", "--ipython_timeout", default=120, type=int)
    input_parser.add_argument("-no_meta_input", "--no_meta_input", help="Do not use meta.json as stimulation input", action="store_true", default=False)

    simulate_parser = sub_parsers.add_parser("simulate")
    simulate_parser.add_argument("path", help="Location of network")
    simulate_parser.add_argument("--networkFile", "--network_file", help="Network file, if not network-synapses.hdf5",
                                 dest="network_file", default=None)
    simulate_parser.add_argument("--inputFile", "--input_file", help="Input hdf5 file (for simulation)",
                                 dest="input_file", default=None)
    simulate_parser.add_argument("--outputFile", "--output_file", help="Output hdf5 file (from simulation)",
                                 dest="output_file", default=None)
    simulate_parser.add_argument("--time", type=float, default=None, help="Duration of simulation in seconds")

    simulate_parser.add_argument("--snudda_data", "--SnuddaData", type=str, default=None, dest="snudda_data",
                                 help="Path to SNUDDA_DATA")
    simulate_parser.add_argument("--simulation_config", type=str, default=None)

    simulate_parser.add_argument("--noVolt", "--novolt", dest="record_volt", action="store_false",
                                 help="Exclude voltage data, to save time and space.")
    simulate_parser.add_argument("-randomseed", "--randomseed", "--seed", default=None, help="Random seed", type=int)

    simulate_parser.add_argument("--disableSyn", "--disableSynapses", action="store_true", dest="disable_synapses", default=None,
                                 help="Disable synapses")

    simulate_parser.add_argument("--disableGJ", "--disableGapJunctions", action="store_true", dest="disable_gj", default=None,
                                 help="Disable gap junctions")

    simulate_parser.add_argument("--mechdir", "--mechDir", "--mech_dir", dest="mech_dir",
                                 help="mechanism directory if not default", default=None)
    simulate_parser.add_argument("--profile", help="Run python cProfile", action="store_true")
    simulate_parser.add_argument("--verbose", action="store_true")
    simulate_parser.add_argument("--export_core_neuron", "--exportCoreNeuron", action="store_true")
    simulate_parser.add_argument("--recordALL", dest="record_all", type=str, default=None)
    simulate_parser.add_argument("--enable_rxd_neuromodulation", dest="enable_rxd_neuromodulation", action="store_true", default=None)
    simulate_parser.add_argument("--disable_rxd_neuromodulation", dest="disable_rxd_neuromodulation", action="store_true", default=None)

    export_parser = sub_parsers.add_parser("export")
    export_parser.add_argument("path", help="Location of network")
    export_parser.add_argument("--inputFile", "--input_file", help="Input hdf5 file (for simulation)",
                               dest="input_file")
    export_parser.add_argument("--networkFile", "--network_file", help="Network file, if not network-synapses.hdf5",
                               dest="network_file")
    export_parser.add_argument("--profile", help="Run python cProfile", action="store_true")
    export_parser.add_argument("--verbose", action="store_true")

    args = parser.parse_args()
    parallel = args.parallel if "parallel" in vars(args) else None
    snudda = Snudda(args.path, parallel=parallel)

    actions = {"init": snudda.init_config_wrapper,
               "import": snudda.import_config_wrapper,
               "create": snudda.create_network_wrapper,
               "place": snudda.place_neurons_wrapper,
               "detect": snudda.detect_synapses_wrapper,
               "prune": snudda.prune_synapses_wrapper,
               "input": snudda.setup_input_wrapper,
               "export": snudda.export_to_SONATA_wrapper,
               "convert": snudda.export_to_SONATA_wrapper,
               "analyse": snudda.analyse,
               "simulate": snudda.simulate_wrapper,
               "help": snudda.help_info}

    if not hasattr(args, 'ipython_profile'):
        args.ipython_profile = None

    print(f"args.ipython_profile = {args.ipython_profile}")

    if args.profile:
        prof_file = f"profile-{args.action}.prof"
        print(f"Saving profile data to: {prof_file}")
        import cProfile
        cProfile.runctx("actions[args.action](args)", None, locals(), filename=prof_file)

        # To analyse profile data:
        import pstats
        from pstats import SortKey
        p = pstats.Stats(prof_file)
        p.strip_dirs().sort_stats(SortKey.CUMULATIVE).print_stats(100)

    else:
        if hasattr(args, "parallel"):
            run_parallel = args.parallel
        else:
            run_parallel = False

        running_neuron = (args.action == "simulate")

        bl = BenchmarkLogging(args.path, parallel_flag=run_parallel, running_neuron=running_neuron,
                              ipython_profile=args.ipython_profile)
        bl.start_timer(args.action)

        # Perform the requested action
        actions[args.action](args)

        bl.stop_timer(args.action)
        bl.write_log()

    snudda.close_log_file()

if __name__ == "__main__":
    snudda_cli()



import sys
import os.path
import h5py
import numpy as np
from mpi4py import MPI  # This must be imported before neuron, to run parallel
from neuron import h  # , gui

from snudda.neurons.NEURON_neuron_extractor import NEURONNeuronExtractor

"""

    Data hierarchy:
    
    SnuddaSaveNetworkRecordings - The Class which contains the data to be saved from the simulation
    
        Each Neuron has an instance of the NeuronRecordings class
        
            Within the NeuronRecordings class, the data can be saved as either:
            
                CompartmentData class, using self.register_compartment_data(**)
                    This contains all data within a neuron compartment (seg)
                
                SynapseData class, using self.register_synapse_data(**)
                    This contains all data within a synapse (point process)
                    
                SpikeData class, using self.register_spike_data(**)
                    This contains all spiking data (events) from a neuron
                    (the only data class which is not a time series)

"""

# TODO:

# Check if we can get synapse row number stored somehow...

# 4. Hur ska spikes sparas? (cell id, tid)
# 5. Gör en generell write funktion som tittar på vilken data som finns och anropar rätt sub-funktioner
# -- Skriv tutorial hur man lägger till en egen recording
#
# Förslag på namn: TimeSeriesData
#                  EventData

# NeuronActivity håller information om alla olika mätningar för en given neuron,
# Simulate har en lista med NeuronActivity
#
# I simulate måste vi:
# För varje neuron som ska mäta något på, skapa en NeuronActivity och lägg i en lista
# För varje measurement vi vill spara (ström, voltage, etc), använd add_measurements
# och sedan


class NeuronRecordings:
    """
        Container class for all recordings associated with a neuron.
    """

    def __init__(self, neuron_id):
        """
        Constructor

        Args:
            neuron_id  (int): Neuron ID
        """

        self.neuron_id = neuron_id

        self.data = dict()

    def register_compartment_data(self, data, data_type, sec_id, sec_x):
        """
        Adds a new measurement.
            
        Args:
            data (neuron.h.Vector): NEURON vector holding recording.
            data_type (str): Name of the tracked data.
            sec_id (int): Section ID
            sec_x (float): Section X (segment location)

        """

        if data_type not in self.data:
            self.data[data_type] = CompartmentData(neuron_id=self.neuron_id, data_type=data_type)

        if self.data[data_type].exist(sec_id=sec_id, sec_x=sec_x):
            print(f"{data_type} already recorded for neuron_id = {self.neuron_id}, {sec_id=}, {sec_x=}, skipping.")
            return

        self.data[data_type].append(data=data, sec_id=sec_id, sec_x=sec_x)

    def register_synapse_data(self, data, data_type, synapse_type, presynaptic_id, sec_id, sec_x, cond):

        if data_type not in self.data:
            self.data[data_type] = SynapseData(neuron_id=self.neuron_id, data_type=data_type)

        self.data[data_type].append(data=data, synapse_type=synapse_type, presynaptic_id=presynaptic_id,
                                    sec_id=sec_id, sec_x=sec_x, cond=cond)

    def register_spike_data(self, data, sec_id, sec_x):

        data_type = "spikes"
        if data_type not in self.data:
            self.data[data_type] = SpikeData(neuron_id=self.neuron_id)

        self.data[data_type].append(data=data, sec_id=sec_id, sec_x=sec_x)


class SynapseData:

    def __init__(self, neuron_id, data_type):
        self.neuron_id = neuron_id
        self.data = []
        self.data_type = data_type
        self.sec_id = []
        self.sec_x = []
        self.synapse_type = []
        self.presynaptic_id = []
        self.cond = []

    def append(self, data, synapse_type, presynaptic_id, sec_id, sec_x, cond):
        self.data.append(data)
        self.synapse_type.append(synapse_type)
        self.sec_id.append(sec_id)
        self.sec_x.append(sec_x)
        self.presynaptic_id.append(presynaptic_id)
        self.cond.append(cond)

    def to_numpy(self):
        """
            Returns:
                (np.ndarray): Data represented as np.ndarrays
        """
        return np.vstack([np.array(d) for d in self.data])


class CompartmentData:
    """
        Container class for recordings from a compartment. 
    """

    def __init__(self, neuron_id, data_type):
        """
        Constructor

        Args:
            neuron_id (int): Neuron ID
            data_type (str): Name of the tracked data.
        """
        self.neuron_id = neuron_id
        self.data_type = data_type
        self.data = []
        self.sec_id = []
        self.sec_x = []

    def exist(self, sec_id, sec_x):
        return any(sec_id == s_id and sec_x == s_x for s_id, s_x in zip(self.sec_id, self.sec_x))

    def append(self, data, sec_id, sec_x):
        # !!! Issue (?): The same compartment can hold several recordings now.

        """
        Appends a recording for a given compartment.

        Args:
            data (neuron.h.Vector): NEURON vector for the recording.
            sec_id (int): Section ID.
            sec_x (float): Section X (segment location). 

        """
        self.data.append(data)

        self.sec_id.append(sec_id)
        self.sec_x.append(sec_x)

    def to_numpy(self):  # Misnomer? (original data is preserved). (Instead "as_ndarray" (like in NEURON)?)
        """
            Returns:
                (np.ndarray): Data represented as np.ndarrays 
        """
        return np.vstack([np.array(d) if d.size() > 0 else np.array([]) for d in self.data])


class SpikeData(CompartmentData):

    def __init__(self, neuron_id):
        super().__init__(neuron_id, data_type="spikes")


class ExtracellularRecordings:

    def __init__(self, region_name, ecs):

        # Needs to identify with its index, and then we also track coordinates
        # we want to avoid duplicate data -- since a node has many coordinates inside it
        # EXAMPLE: https://neuron.yale.edu/neuron/docs/extracellular-diffusion

        self.ecs = ecs   # remove?
        self.region_name = region_name
        self.data = dict()
        self.coords = dict()

        self.xyz_min = np.array((self.ecs[self.region_name].compartments["ecs"]._xlo,
                                 self.ecs[self.region_name].compartments["ecs"]._ylo,
                                 self.ecs[self.region_name].compartments["ecs"]._zlo))

        self.xyz_max = np.array((self.ecs[self.region_name].compartments["ecs"]._xhi,
                                 self.ecs[self.region_name].compartments["ecs"]._yhi,
                                 self.ecs[self.region_name].compartments["ecs"]._zhi))

        self.n_voxels = np.array((self.ecs[self.region_name].compartments["ecs"]._nx,
                                  self.ecs[self.region_name].compartments["ecs"]._ny,
                                  self.ecs[self.region_name].compartments["ecs"]._nz))

    def get_coords(self, index_ijk):
        return np.vstack([self.coords[tuple(ijk)] for ijk in index_ijk])

    def register_extracellular_data(self, data, data_type, index_ijk):

        if data_type not in self.data:
            self.data[data_type] = ExtracellularNodeData(data_type=data_type)

        if index_ijk not in self.coords:
            delta_xyz = (self.xyz_max - self.xyz_min) / self.n_voxels
            self.coords[index_ijk] = self.xyz_min + delta_xyz * index_ijk

        self.data[data_type].append(data=data, index_ijk=index_ijk)

        # TODO: This object tracks the data in one "node",
        # The node knows its i,j,
        # It will have a dictionary, and there we need to know the species type, and the location
        # we will use i, j, k for index (but also store the centre of the node)
        # This node needs to know its own region, species, and location (x,y,z)
        #  We need to save the extracellular data, based on region, species, loc

# TODO: Check if RxD uses JIT, as they claim in Omar Awile et al, 2022.


class ExtracellularNodeData:

    def __init__(self, data_type):
        self.data_type = data_type
        self.data = dict()

    def append(self, data, index_ijk, unsafe=False):
        if not unsafe and index_ijk in self.data:
            raise ValueError(f"Index {index_ijk} already exists in the data structure.")

        self.data[index_ijk] = data

    def sort(self):
        sorted_items = sorted(self.data.items(), key=lambda x: x[0])
        self.data = dict(sorted_items)

    def to_numpy(self):
        return np.vstack([np.array(d) for d in self.data.values()])

    def get_data(self):
        self.sort()

        data = self.to_numpy()
        index = np.array(list(self.data.keys()))

        return index, data


class SnuddaSaveNetworkRecordings:

    # TODO: Add saving of simulation_config file (and experiment_config_file for pair recording)

    def __init__(self, output_file, network_data=None, sample_dt=None, node_id=0):

        # Only do this check on the first node
        if node_id == 0 and not self.check_file_available(output_file):
            sys.exit(-1)

        self.output_file = output_file
        self.network_data = network_data
        self.header_exists = False
        self.neuron_activities = dict()
        self.extracellular_data = dict()
        self.time = None
        self.sample_dt = sample_dt

        self.units = dict()

        self.pc = h.ParallelContext()

        self.save_geometry = False
        self.simulation = None

    def include_geometry(self, simulation):
        self.simulation = simulation
        self.save_geometry = True

    def set_new_output_file(self, output_file):
        self.output_file = output_file
        self.header_exists = False

    def add_unit(self, data_type, target_unit, conversion_factor):
        # Units reference chart: https://www.neuron.yale.edu/neuron/static/docs/units/unitchart.html
        self.units[data_type] = {"unit": target_unit, "conversion_factor": conversion_factor}

    def get_conversion(self, data_type):
        if data_type in self.units:
            return self.units[data_type]["conversion_factor"]
        else:
            return 1.0

    def register_compartment_data(self, data_type, neuron_id, data, sec_id, sec_x):
        if neuron_id not in self.neuron_activities:
            self.neuron_activities[neuron_id] = NeuronRecordings(neuron_id)

        self.neuron_activities[neuron_id].register_compartment_data(data=data, data_type=data_type,
                                                                    sec_id=sec_id, sec_x=sec_x)

    def register_synapse_data(self, data_type, neuron_id, data, synapse_type, presynaptic_id, sec_id, sec_x, cond):
        if neuron_id not in self.neuron_activities:
            self.neuron_activities[neuron_id] = NeuronRecordings(neuron_id)

        self.neuron_activities[neuron_id].register_synapse_data(data=data, data_type=data_type,
                                                                synapse_type=synapse_type,
                                                                presynaptic_id=presynaptic_id,
                                                                sec_id=sec_id, sec_x=sec_x, cond=cond)

    def register_spike_data(self, neuron_id, data, sec_id, sec_x):
        if neuron_id not in self.neuron_activities:
            self.neuron_activities[neuron_id] = NeuronRecordings(neuron_id)

        self.neuron_activities[neuron_id].register_spike_data(data=data, sec_id=sec_id, sec_x=sec_x)

    def register_extracellular_data(self, region_name, data_type, data, ecs, index_ijk):

        if region_name not in self.extracellular_data:
            self.extracellular_data[region_name] = ExtracellularRecordings(region_name=region_name, ecs=ecs)

        self.extracellular_data[region_name].register_extracellular_data(data=data, data_type=data_type,
                                                                         index_ijk=index_ijk)

    def register_time(self, time):
        self.time = time

    @staticmethod
    def spike_sort(t_spikes, id_spikes):

        spike_count = dict()

        for t, idx in zip(t_spikes, id_spikes):
            if idx in spike_count:
                spike_count[idx] += 1
            else:
                spike_count[idx] = 1

        spikes = dict()
        spike_ctr = dict()

        for idx, num in spike_count.items():
            spikes[idx] = np.full((num,), np.nan)  # Initialise to NaN to catch any unassigned values
            spike_ctr[idx] = 0

        for t, idx in zip(t_spikes, id_spikes):
            spikes[idx][spike_ctr[idx]] = t
            spike_ctr[idx] += 1

        # Internal consistency
        for idx in spikes:
            assert len(spikes[idx]) == 0 or not np.isnan(spikes[idx][-1])
            assert (idx in spike_ctr and spikes[idx].shape[0] == spike_ctr[idx]) \
                   or (idx not in spike_ctr and len(spikes[idx]) == 0)

        return spikes

    def write_string_meta_data(self, group, name):
        string_data = [x[name] if x[name] is not None else "" for x in self.network_data["neurons"]]
        max_len = max(1, max([len(x) for x in string_data]))
        str_type = f"S{max_len}"
        group.create_dataset(name, (len(string_data),), str_type, string_data, compression="gzip")

    def check_file_available(self, file_name):

        if os.path.isfile(file_name):
            # Try to open and close file, to make sure it is available
            try:
                f = h5py.File(file_name, "w")
                f.close()
            except BlockingIOError as e:
                print(f"Unable to create file {file_name}. Is some other program using the file?")
                return False

        return True

    def write_header(self):

        self.pc.barrier()

        if self.header_exists:
            return

        if int(self.pc.id()) == 0:

            if not os.path.isdir(os.path.dirname(self.output_file)):
                os.mkdir(os.path.dirname(self.output_file))

            print(f"Writing network output to {self.output_file}")

            try:
                out_file = h5py.File(self.output_file, "w")
            except Exception as e:

                print(e)
                print(f"Trying to recover, and save to different file name.")

                ctr = 1
                temp_name = f"{self.output_file}-{ctr}"

                while os.path.isfile(temp_name):
                    print(f"File exists: {temp_name}")
                    ctr += 1
                    temp_name = f"{self.output_file}-{ctr}"

                print(f"\n!!! Unable to create {self.output_file} (file locked?), using {temp_name} instead\n\n")
                self.output_file = temp_name

                out_file = h5py.File(temp_name, "w")

            meta_data = out_file.create_group("meta_data")

            meta_data.create_dataset("network_file", data=self.network_data["network_file"])
            out_file.create_group("neurons")

            if self.network_data:
                neuron_id = np.array([x["neuron_id"] for x in self.network_data["neurons"]])
                meta_data.create_dataset("id", data=neuron_id)

                for name in ["name", "type", "morphology", "parameter_key", "morphology_key", "modulation_key"]:
                    self.write_string_meta_data(group=meta_data, name=name)

                meta_data.create_dataset("population_unit", data=self.network_data["population_unit"], compression="gzip")
                meta_data.create_dataset("position", data=self.network_data["neuron_positions"], compression="gzip")
                meta_data.create_dataset("virtual_neuron", compression="gzip",
                                         data=[n["virtual_neuron"] for n in self.network_data["neurons"]])

                # TODO: Also write extraAxons info to file

            out_file.close()

        self.header_exists = True

        self.pc.barrier()

    def get_sample_step(self):

        if self.sample_dt is None:
            return None
        else:
            converted_time = np.array(self.time) * self.get_conversion("time")
            dt = converted_time[1] - converted_time[0]
            sample_step = int(np.round(self.sample_dt / dt))

            if sample_step == 0:
                raise ValueError(f"sample_step can not be zero. {self.sample_dt = } should be equal or larger than {dt = }")

            return sample_step

    def write(self):

        self.write_header()

        sample_step = self.get_sample_step()

        if int(self.pc.id()) == 0:

            # TODO: We need to save max time even if we do not save the soma voltage(!)
            out_file = h5py.File(self.output_file, "a")

            if "time" not in out_file and self.time is not None:
                print(f"Using sample dt = {self.sample_dt} (sample step size {sample_step})")
                try:
                    out_file.create_dataset("time", data=np.array(self.time)[::sample_step] * self.get_conversion("time"))
                    out_file.close()
                except:
                    import traceback
                    print(f"self.time={self.time}")
                    print(f"{traceback.format_exc()}", flush=True)
                    import pdb
                    pdb.set_trace()

        for i in range(int(self.pc.nhost())):

            self.pc.barrier()

            if int(self.pc.id()) == i:
                print(f"Worker {i+1}/{int(self.pc.nhost())} writing data to {self.output_file}")

                out_file = h5py.File(self.output_file, "a")

                # Make sure time data is written to file, in case node 0 only has virtual neurons
                if "time" not in out_file and self.time is not None:
                    print(f"Using sample dt = {self.sample_dt} (sample step size {sample_step})")
                    try:
                        out_file.create_dataset("time",
                                                data=np.array(self.time)[::sample_step] * self.get_conversion("time"))
                        out_file.close()
                    except:
                        import traceback
                        print(f"self.time={self.time}")
                        print(f"{traceback.format_exc()}", flush=True)
                        import pdb
                        pdb.set_trace()

                for na in self.neuron_activities.values():
                    neuron_id_str = str(na.neuron_id)

                    if neuron_id_str not in out_file["neurons"]:
                        out_file["neurons"].create_group(neuron_id_str)

                    for m in na.data.values():

                        conversion_factor = self.get_conversion(m.data_type)

                        if isinstance(m, SpikeData):
                            # Spike data is not a time series, and should never be downsampled
                            data_sample_step = None
                        else:
                            data_sample_step = sample_step

                        data_set = out_file["neurons"][neuron_id_str].create_dataset(m.data_type,
                                                                                     data=m.to_numpy()[:, ::data_sample_step]*conversion_factor,
                                                                                     compression="gzip")

                        data_set.attrs["sec_id"] = np.array(m.sec_id)
                        data_set.attrs["sec_x"] = np.array(m.sec_x)

                        if isinstance(m, SynapseData):
                            try:
                                data_set.attrs["synapse_type"] = np.array(m.synapse_type)
                                data_set.attrs["presynaptic_id"] = np.array(m.presynaptic_id)
                                data_set.attrs["cond"] = np.array(m.cond)
                            except Exception as e:
                                import traceback
                                print(traceback.format_exc())
                                import pdb
                                pdb.set_trace()

                if int(self.pc.id()) == 0:
                    # RxD computes extracellular diffusion INDEPENDENTLY on ALL nodes (sigh!)
                    # thus we need to save only data from node 0

                    ecs_group = out_file.create_group("ecs")

                    for region_name, region_data in self.extracellular_data.items():
                        region_group = ecs_group.create_group(region_name)

                        for data_type, extracellular_data in region_data.data.items():
                            ecs_data_group = region_group.create_group(data_type)

                            index, data = extracellular_data.get_data()
                            coords = region_data.get_coords(index)

                            ecs_data_group.create_dataset("index", data=index)
                            ecs_data_group.create_dataset("xyz", data=coords)
                            ecs_data_group.create_dataset("data", data=data)

                if self.save_geometry:
                    nne = NEURONNeuronExtractor(sim=self.simulation)
                    nne.write_all_coordinates_to_hdf5(hdf5_file=out_file)

                out_file.close()

        self.pc.barrier()

